
#include "Node.hpp"

namespace SPN {

	const double GaussianNode::logSqrt2Pi = log(sqrt(2 * 3.141592653589793238462643383279));

	GaussianNode::GaussianNode() : Node(GAUSSIAN_NODE) {
		means_.clear();
		sigmas_.clear();		
		mode_ = PDF_MODE;
	}

	GaussianNode::~GaussianNode() {
	}

	void GaussianNode::eval() {
		logVal_ = 0.0;
		size_t numChildren = getNumChildren();
		if (numChildren == 0)
			return;

		if (mode_ == PDF_MODE) {
			for (unsigned k = 0; k < numChildren; k++) {
				Node* child = getChild(k);

				// when child is a val node, whose value is set to nan -> interpret as missing value
				if ((child->getType() == VAL_NODE) && (isnan(child->getVal())))
					continue;

				NodeId childId = child->getId();
				logVal_ += (-log(sigmas_[childId]) - logSqrt2Pi) -
					(0.5 * pow(((child->getVal() - means_[childId]) / sigmas_[childId]), 2));

			}
		}
		else if (mode_ == CDF_MODE) {
			for (unsigned k = 0; k < numChildren; k++) {
				Node* child = getChild(k);

				// when child is a val node, whose value is set to nan -> interpret as missing value
				if ((child->getType() == VAL_NODE) && (isnan(child->getVal())))
					continue;

				NodeId childId = child->getId();
				double tmp = 0.5 * (1 + erf((child->getVal() - means_[childId]) / sqrt(2 * pow(sigmas_[childId], 2))));
				if (tmp > 0)
					logVal_ += log(tmp);
				else {
					logVal_ = -inf;
					break;
				}
			}
		}
		else
			throw std::runtime_error("GaussianNode::eval(): unknown mode");
	}

	double GaussianNode::getLog_Dval_Dchild(Node *child) {
		if (!isChild(child))
			throw std::runtime_error("GaussianNode::getLog_Dval_Dchild(Node *child): child node not found");

		if (isnan(child->getVal()))
			return nan;

		double logDval;
		NodeId childId = child->getId();

		if (mode_ == PDF_MODE)
			logDval = (-log(sigmas_[childId]) - logSqrt2Pi) -
			(0.5 * pow(((child->getVal() - means_[childId]) / sigmas_[childId]), 2));
		else {
			logDval = 0.5 * (1 + erf((child->getVal() - means_[childId]) / sqrt(2 * pow(sigmas_[childId], 2))));
			if (logDval > 0)
				logDval = log(logDval);
			else
				logDval = -inf;
		}
		return logDval;
	}

	void GaussianNode::connectChild(Node *child) {
		connectChild(child, 0.0, 1.0);
	}

	void GaussianNode::connectChild(Node *child, double mean, double sigma) {
		if ((isChild(child)) || (child->getId() == getId()))
			return;

		Node::connectChild(child);
		setMean(child, mean);
		setSigma(child, sigma);		
	}

	void GaussianNode::disconnectChild(Node *child) {
		Node::disconnectChild(child);
		NodeId childId = child->getId();
		means_.erase(childId);
		sigmas_.erase(childId);		
	}

	void GaussianNode::setMean(Node *child, double mean) {
		if (!isChild(child))
			throw std::runtime_error("error in GaussianNode::setMean(Node *child, double mean): not a child.");
		means_[child->getId()] = mean;
	}

	double GaussianNode::getMean(Node *child) {
		if (!isChild(child))
			throw std::runtime_error("error in GaussianNode::getMean(Node *child): not a child.");
		return means_[child->getId()];
	}

	void GaussianNode::setMeansRand(double minVal, double maxVal) {
		for (unsigned k = 0; k < getNumChildren(); k++) {
			setMean(getChild(k), minVal + (maxVal - minVal) * (((double)rand()) / ((double)RAND_MAX)));
		}
	}

	void GaussianNode::setSigma(Node *child, double sigma) {
		if (!isChild(child))
			throw std::runtime_error("error in GaussianNode::setSigma(Node *child, double sigma): not a child.");
		if (sigma <= 0)
			throw std::runtime_error("error in GaussianNode::setSigma(Node *child, double sigma): "
			"sigma has to be larger than 0.");
		sigmas_[child->getId()] = sigma;
	}

	double GaussianNode::getSigma(Node *child) {
		if (!isChild(child))
			throw std::runtime_error("error in GaussianNode::getSigma(Node *child): not a child.");
		return sigmas_[child->getId()];
	}

	void GaussianNode::setSigmasRand(double minVal, double maxVal) {
		for (unsigned k = 0; k < getNumChildren(); k++) {
			setSigma(getChild(k), minVal + (maxVal - minVal) * (((double)rand()) / ((double)RAND_MAX)));
		}
	}
	
	void GaussianNode::setMode(int mode) {
		mode_ = mode;
	}

}
