
#include "Node.hpp"

namespace SPN {

	double IndicatorNode::tol_ = 1e-12;

	IndicatorNode::IndicatorNode() : Node(INDICATOR_NODE) {
		logVal_ = 0.0;
		indValues_.clear();
		clampVal_ = -1;
	}

	IndicatorNode::~IndicatorNode() {
	}

	void IndicatorNode::eval() {
		logVal_ = 0.0;
		size_t numChildren = getNumChildren();

		for (unsigned k = 0; k < numChildren; k++) {
			Node* child = getChild(k);

			// when child is a val node, whose value is set to nan -> interpret as missing value
			if ((child->getType() == VAL_NODE) && (isnan(child->getVal())))
				continue;

			if (fabs(indValues_[child->getId()] - child->getVal()) > tol_) {
				logVal_ = -inf;
				break;
			}
		}
	}

	double IndicatorNode::getLog_Dval_Dchild(Node *child) {
		return logVal_;
	}

	void IndicatorNode::connectChild(Node *child) {
		connectChild(child, 0.0);
	}

	void IndicatorNode::connectChild(Node *child, double indValue) {
		if ((isChild(child)) || (child->getId() == getId()))
			return;

		Node::connectChild(child);
		setIndValue(child, indValue);
	}

	void IndicatorNode::disconnectChild(Node *child) {
		Node::disconnectChild(child);
		indValues_.erase(child->getId());
	}

	double IndicatorNode::getLogVal() const {
		if (clampVal_ < 0)
			return logVal_;
		else {
			if (clampVal_ == 0)
				return -inf;
			else
				return 0.0;
		}
	}

	double IndicatorNode::getVal() const {
		if (clampVal_ < 0) {
			if (logVal_ == -inf)
				return 0.0;
			else if (logVal_ == 0.0)
				return 1.0;
			else
				throw std::runtime_error("IndicatorNode::getVal(): inconsistent value");
		}
		else {
			if (clampVal_ == 0)
				return 0.0;
			else
				return 1.0;
		}
	}

	void IndicatorNode::setIndValue(Node *child, double indValue) {
		if (!isChild(child))
			throw std::runtime_error("error in IndicatorNode::setIndValue(Node *child, double indValue): not a child.");

		indValues_[child->getId()] = indValue;
	}

	double IndicatorNode::getIndValue(Node *child) {
		if (!isChild(child))
			throw std::runtime_error("error in IndicatorNode::getIndValue(Node *child): not a child.");

		return indValues_[child->getId()];
	}

	void IndicatorNode::clamp(bool clampVal) {
		if (clampVal)
			clampVal_ = 1;
		else
			clampVal_ = 0;
	}

	void IndicatorNode::unclamp() {
		clampVal_ = -1;
	}

	void IndicatorNode::setTolerance(double tol) {
		tol_ = tol;
	}

	double IndicatorNode::getTolerance() {
		return tol_;
	}

}

