
#include "Layer.hpp"

namespace SPN {

	LayerId Layer::idCounter_ = 0;

	Layer::Layer(LayerType type) {
		type_ = type;
		id_ = idCounter_++;		
		nodes_.clear();
		nodeSet_.clear();
	}

	Layer::~Layer() {		
	}

	void Layer::eval() {
#if MULTITHREADING == 1
		omp_set_num_threads(NUM_THREADS);
#pragma omp parallel for
		for (long int k = 0; k < (long int) nodes_.size(); k++)
			nodes_[k]->eval();
#else
		for (unsigned k = 0; k < nodes_.size(); k++)
			nodes_[k]->eval();
#endif
	}

	void Layer::backprop() {
#if MULTITHREADING == 1
		omp_set_num_threads(NUM_THREADS);
#pragma omp parallel for
		for (long int k = 0; k < (long int) nodes_.size(); k++)
			nodes_[k]->backprop();
#else
		for (unsigned k = 0; k < nodes_.size(); k++)
			nodes_[k]->backprop();
#endif
	}

	void Layer::insertNode(Node *node) {
		insertNodes(NodeVec(1, node));
	}

	void Layer::insertNodes(NodeVec nodes) {
		for (size_t k = 0; k < nodes.size(); k++) {
			if (((type_ == Layer::SUM_LAYER) && (nodes[k]->getType() != Node::SUM_NODE)) ||
				((type_ == Layer::PROD_LAYER) && (nodes[k]->getType() != Node::PROD_NODE)) ||
				((type_ == Layer::INDICATOR_LAYER) && (nodes[k]->getType() != Node::INDICATOR_NODE)) ||
				((type_ == Layer::VAL_LAYER) && (nodes[k]->getType() != Node::VAL_NODE)) ||
				((type_ == Layer::GAUSSIAN_LAYER) && (nodes[k]->getType() != Node::GAUSSIAN_NODE)))
				throw std::runtime_error("Layer::insertNodes(NodeVec nodes): wrong type");
			if (nodeSet_.count(nodes[k]->getId()))
				continue;
			nodeSet_.insert(nodes[k]->getId());
			nodes_.push_back(nodes[k]);
		}
	}

	void Layer::removeNode(Node *node) {
		nodes_.erase(nodes_.begin() + getNodeIdx(node));

		if (nodeSet_.count(node->getId()) == 0)
			throw std::runtime_error("Layer::removeNode(Node *node): inconsistent nodeIndMap.");
		nodeSet_.erase(node->getId());
	}

	void Layer::removeNodes(NodeVec nodes) {
		for (size_t k = 0; k < nodes.size(); k++)
			removeNode(nodes[k]);
	}

	size_t Layer::getNumNodes() {
		return nodes_.size();
	}

	NodeVec Layer::getNodes() {
		return nodes_;
	}

	Node* Layer::getNode(size_t idx) {
		if (idx >= nodes_.size())
			throw std::runtime_error("error in Layer::getNode(size_t idx): idx out of bounds.");
		return (nodes_[idx]);
	}

	size_t Layer::getNodeIdx(Node* node) {
		NodeId nodeId = node->getId();
		size_t retidx = 0;
		bool found = false;

		for (unsigned k = 0; k < nodes_.size(); k++) {
			if (nodes_[k]->getId() == nodeId) {
				retidx = k;
				found = true;
				break;
			}
		}

		if (!found)
			throw std::runtime_error("error in Layer::getNodeIdx(Node* node): "
			"layer does not contain node.");

		return retidx;
	}

	bool Layer::containsNode(Node *node) {
		return (nodeSet_.count(node->getId()) != 0);
	}

	int Layer::getType() const {
		return type_;
	}

	LayerId Layer::getId() const {
		return id_;
	}

	Node* Layer::operator()(size_t idx) {
		return getNode(idx);
	}
}


