
#include "Network.hpp"
#include <iostream>

namespace SPN {

	NetworkId Network::idCounter_ = 0;

	Network::Network() {
		id_ = idCounter_++;
	}

	Network::~Network() {
	}

	void Network::insertLayer(Layer* layer, size_t idx) {
		if (containsLayer(layer))
			return;

		if (idx >= getNumLayers())
			layers_.push_back(layer);
		else
			layers_.insert(layers_.begin() + idx, layer);
	}

	void Network::addLayer(Layer* layer) {
		insertLayer(layer, getNumLayers());
	}

	void Network::removeLayer(Layer *layer) {
		if (!containsLayer(layer))
			throw std::runtime_error("Network::removeLayer(Layer *layer):tried to remove non-containing layer.");
		layers_.erase(layers_.begin() + getLayerIdx(layer));
	}

	size_t Network::getNumLayers() {
		return layers_.size();
	}

	LayerVec Network::getLayers() {
		return layers_;
	}

	Layer* Network::getLayer(size_t layerIdx) {
		if (layerIdx >= layers_.size())
			throw std::runtime_error("error in Network::getLayer(unsigned idx): idx out of bounds.");
		return layers_[layerIdx];
	}

	Layer* Network::getTopLayer() {
		if (layers_.size() == 0)
			return NULL;
		return layers_[layers_.size()-1];
	}

	size_t Network::getLayerIdx(Layer* layer) {
		unsigned retidx = 0;
		bool found = false;
		LayerId layerId = layer->getId();

		for (unsigned k = 0; k < layers_.size(); k++) {
			if (layers_[k]->getId() == layerId) {
				retidx = k;
				found = true;
				break;
			}
		}

		if (!found)
			throw std::runtime_error("error in Network::getLayerIdx(Layer* layer): network does not contain layer.");

		return retidx;
	}

	Node* Network::getNode(size_t layerIdx, size_t idx) {
		return getLayer(layerIdx)->getNode(idx);
	}

	size_t Network::getLayerIdxOfNode(Node *node) {
		for (unsigned k = 0; k < getNumLayers(); k++)
			if (getLayer(k)->containsNode(node))
				return k;
		throw std::runtime_error("error in Network::getLayerIdxOfNode(Node *node): "
			"network does not contain node.");
	}

	bool Network::containsLayer(Layer *layer) {
		bool found = false;
		LayerId layerId = layer->getId();

		for (unsigned k = 0; k < layers_.size(); k++) {
			if (layers_[k]->getId() == layerId) {
				found = true;
				break;
			}
		}

		return found;
	}

	void Network::setVals(const DoubleVec &values) {
		if (getNumLayers() == 0)
			throw std::runtime_error("error in Network::setVals(const DoubleVec &values): Network has no layers.");

		if (layers_[0]->getType() != Layer::VAL_LAYER)
			throw std::runtime_error("error in Network::setVals(const DoubleVec &values): layer 0 has to be of type Layer::VAL_LAYER.");

		((ValLayer*)layers_[0])->setVals(values);
	}

	DoubleVec Network::getVals() {
		if (getNumLayers() == 0)
			throw std::runtime_error("error in Network::getVals(): Network has no layers.");

		if (layers_[0]->getType() != Layer::VAL_LAYER)
			throw std::runtime_error("error in Network::getVals(): layer 0 has to be of type Layer::VAL_LAYER.");

		return ((ValLayer*)layers_[0])->getVals();
	}

	void Network::eval() {
		for (unsigned k = 0; k < layers_.size(); k++)
			layers_[k]->eval();
	}

	void Network::backprop() {
		for (unsigned k = 0; k < layers_.size(); k++)
			layers_[layers_.size() - k - 1]->backprop();
	}

	double Network::loglikelihood(DoubleVec &data) {
		if (getTopLayer()->getNumNodes() != 1)
			throw std::runtime_error("Network::loglikelihood: top layer must have exactly one node");
		setVals(data);
		eval();
		return getTopLayer()->getNode(0)->getLogVal();
	}

	void Network::MPEinference() {
		if (getLayer(getNumLayers() - 1)->getType() != Layer::SUM_LAYER)
			throw std::runtime_error("Network::MPEinference(): last layer has to be a sum layer");

		if (getLayer(getNumLayers() - 1)->getNumNodes() != 1)
			throw std::runtime_error("Network::MPEinference(): last layer must contain exactly one node");

		eval();
		backprop();

		std::vector<SumNode* > maxSumNodes;

		maxSumNodes.clear();
		maxSumNodes.push_back((SumNode*)getLayer(getNumLayers() - 1)->getNode(0));

		for (unsigned k = 0; k < maxSumNodes.size(); k++) {
			SumNode* curSumNode = maxSumNodes[k];
			std::vector<ProdNode*> prodNodeOptions;
			prodNodeOptions.clear();
			double maxVal = -inf;

			for (unsigned l = 0; l < curSumNode->getNumChildren(); l++) {
				if (curSumNode->getChild(l)->getType() != Node::PROD_NODE)
					throw std::runtime_error("Network::MPEinference(): children of sum nodes have to be product nodes");

				double curVal = curSumNode->getChild(l)->getLogVal() + curSumNode->getLogWeight(curSumNode->getChild(l));
				if ((l == 0) || (curVal > maxVal)) {
					maxVal = curVal;
					prodNodeOptions.clear();
				}
				if (curVal == maxVal)
					prodNodeOptions.push_back((ProdNode*)curSumNode->getChild(l));
			}

			std::queue<ProdNode*> prodQ;
			prodQ.push(prodNodeOptions[rand() % prodNodeOptions.size()]);

			while (!prodQ.empty()) {
				ProdNode* prodNode = prodQ.front();
				prodQ.pop();

				for (unsigned l = 0; l < prodNode->getNumChildren(); l++) {
					if (prodNode->getChild(l)->getType() == Node::SUM_NODE)
						maxSumNodes.push_back((SumNode*)prodNode->getChild(l));
					else if (prodNode->getChild(l)->getType() == Node::GAUSSIAN_NODE) {
						GaussianNode* gaussNode = (GaussianNode*)prodNode->getChild(l);
						for (unsigned m = 0; m < gaussNode->getNumChildren(); m++) {
							if (gaussNode->getChild(m)->getType() != Node::VAL_NODE)
								throw std::runtime_error("error: child of gauss has to be val node.");
							ValNode* valNode = (ValNode*)(gaussNode->getChild(m));
							if (isnan(valNode->getVal()))
								valNode->setVal(gaussNode->getMean(valNode));
						}
					}
					else if (prodNode->getChild(l)->getType() == Node::PROD_NODE)
						prodQ.push((ProdNode*)prodNode->getChild(l));
					else
						throw std::runtime_error("MPEInference: unexpected node type.");
				}
			}
		}
	}
	
	void Network::setSumMode() {
		for (unsigned k = 0; k < layers_.size(); k++)
			if (layers_[k]->getType() == Layer::SUM_LAYER)
				for (unsigned l = 0; l < layers_[k]->getNumNodes(); l++)
					((SumNode*)layers_[k]->getNode(l))->setSumMode();
	}

	void Network::setMaxMode() {
		for (unsigned k = 0; k < layers_.size(); k++)
			if (layers_[k]->getType() == Layer::SUM_LAYER)
				for (unsigned l = 0; l < layers_[k]->getNumNodes(); l++)
					((SumNode*)layers_[k]->getNode(l))->setMaxMode();
	}

	void Network::setGaussianMeansRand(double minVal, double maxVal) {
		for (unsigned k = 0; k < layers_.size(); k++)
			if (layers_[k]->getType() == Layer::GAUSSIAN_LAYER)
				((GaussianLayer*)layers_[k])->setMeansRand(minVal, maxVal);
	}

	void Network::setGaussianSigmasRand(double minVal, double maxVal) {
		for (unsigned k = 0; k < layers_.size(); k++)
			if (layers_[k]->getType() == Layer::GAUSSIAN_LAYER)
				((GaussianLayer*)layers_[k])->setSigmasRand(minVal, maxVal);
	}

	void Network::setGaussianMeansUniform(double meanVal) {
		for (unsigned k = 0; k < layers_.size(); k++)
			if (layers_[k]->getType() == Layer::GAUSSIAN_LAYER)
				((GaussianLayer*)layers_[k])->setMeansUniform(meanVal);
	}

	void Network::setGaussianSigmasUniform(double sigmaVal) {
		for (unsigned k = 0; k < layers_.size(); k++)
			if (layers_[k]->getType() == Layer::GAUSSIAN_LAYER)
				((GaussianLayer*)layers_[k])->setSigmasUniform(sigmaVal);
	}

	void Network::initWeightsRand() {
		for (unsigned k = 0; k < layers_.size(); k++)
			if (layers_[k]->getType() == Layer::SUM_LAYER)
				((SumLayer*)layers_[k])->initWeightsRand();
	}

	void Network::clipWeights(double threshold) {
		for (unsigned k = 0; k < getNumLayers(); k++) {
			if (getLayer(k)->getType() == Layer::SUM_LAYER) {
				for (unsigned l = 0; l < getLayer(k)->getNumNodes(); l++) {
					SumNode* sumNode = (SumNode*)getNode(k, l);

					for (unsigned m = 0; m < sumNode->getNumChildren(); m++)
						if (sumNode->getWeight(sumNode->getChild(m)) <= threshold)
							sumNode->setWeight(sumNode->getChild(m), 0.0);

				}
			}
		}
	}

	void Network::removeSingletonParents(NodeVec &removedNodes, LayerVec &removedLayers) {
		removedNodes.clear();
		removedLayers.clear();

		for (size_t k = 0; k < getNumLayers(); k++) {
			NodeVec removed;
			for (size_t l = 0; l < getLayer(k)->getNumNodes(); l++) {
				Node* n = getLayer(k)->getNode(l);
				if ((n->getType() != Node::SUM_NODE) && (n->getType() != Node::PROD_NODE))
					continue;
				if (n->getNumChildren() == 1) {
					Node* child = n->getChild(0);

					//if (n->getType() == Node::SUM_NODE)
					//	std::cout << "sum " << ((SumNode*)n)->getLogWeight(child) << std::endl;
					//if (n->getType() == Node::PROD_NODE)
					//	std::cout << "prod " << std::endl;

					for (size_t p = 0; p < n->getNumParents(); p++) {
						Node* parent = n->getParent(p);
						if (parent->getType() == Node::SUM_NODE) {
							parent->connectChild(child);
							((SumNode*)parent)->setLogWeight(child, ((SumNode*)parent)->getLogWeight(n));
						}
						else if (parent->getType() == Node::PROD_NODE)
							parent->connectChild(child);
						else
							throw std::runtime_error("Network::removeSingletonParents(): unexpected node type");
					}
					n->isolate();
					removed.push_back(n);
				}
			}
			getLayer(k)->removeNodes(removed);
			removedNodes.insert(removedNodes.end(), removed.begin(), removed.end());
		}

		for (size_t k = 0; k < getNumLayers(); k++)
			if (getLayer(k)->getNumNodes() == 0)
				removedLayers.push_back(getLayer(k));

		for (size_t k = 0; k < removedLayers.size(); k++)
			removeLayer(removedLayers[k]);
	}

	void Network::disconnectZeroChildren() {
		std::vector<Node*> childrenToDisconnect;

		for (unsigned k = 0; k < getNumLayers(); k++) {
			if (getLayer(k)->getType() == Layer::SUM_LAYER) {
				for (unsigned l = 0; l < getLayer(k)->getNumNodes(); l++) {
					SumNode* sumNode = (SumNode*)getNode(k, l);
					childrenToDisconnect.clear();

					for (unsigned m = 0; m < sumNode->getNumChildren(); m++)
						if (sumNode->getWeight(sumNode->getChild(m)) == 0.0)
							childrenToDisconnect.push_back(sumNode->getChild(m));

					for (unsigned m = 0; m < childrenToDisconnect.size(); m++)
						sumNode->disconnectChild(childrenToDisconnect[m]);
				}
			}
		}
	}

	void Network::normalizeWeights() {
		for (unsigned k = 0; k < getNumLayers(); k++) {
			if (getLayer(k)->getType() == Layer::SUM_LAYER) {
				for (unsigned l = 0; l < getLayer(k)->getNumNodes(); l++) {
					SumNode* sumNode = (SumNode*)getNode(k, l);
					double sum = 0.0;
					for (unsigned m = 0; m < sumNode->getNumChildren(); m++)
						sum += sumNode->getWeight(sumNode->getChild(m));
					if (sum == 0.0)
						return;
					for (unsigned m = 0; m < sumNode->getNumChildren(); m++)
						sumNode->setWeight(sumNode->getChild(m), sumNode->getWeight(sumNode->getChild(m)) / sum);
				}
			}
		}
	}

	void Network::normalizeWeightsProbInvariant() {

		if (getTopLayer()->getNumNodes() != 1)
			throw std::runtime_error("Network::normalizeWeightsProbInvariant(): last layer contains more than one node");

		NodeVec nodes,order;
		listDescendants(getTopLayer()->getNode(0), nodes);
		topOrder(nodes, order, getTopLayer()->getNode(0));

		UMap<NodeId, double> logvals;

		for (size_t k = 0; k < order.size(); k++) {
			size_t idx = order.size() - k - 1;
			Node* n = order[idx];
			if ((n->getType() == Node::GAUSSIAN_NODE) || (n->getType() == Node::INDICATOR_NODE))
				logvals[n->getId()] = 0.0;
			else if (n->getType() == Node::PROD_NODE) {
				double v = 0.0;
				for (size_t c = 0; c < n->getNumChildren(); n++)
					v += logvals.at(n->getChild(c)->getId());
				logvals[n->getId()] = v;
			}
			else if (n->getType() == Node::SUM_NODE) {
				DoubleVec cv;
				for (size_t c = 0; c < n->getNumChildren(); n++)
					cv.push_back(((SumNode*)n)->getLogWeight(n->getChild(c)) + logvals.at(n->getChild(c)->getId()));
				double lse = logSumExp(cv);
				logvals[n->getId()] = lse;
				for (size_t c = 0; c < n->getNumChildren(); n++)
					((SumNode*)n)->setLogWeight(n->getChild(c), ((SumNode*)n)->getLogWeight(n->getChild(c)) - lse);
			}
		}
	}

	void Network::clear() {
		layers_.clear();
	}

	size_t Network::getNumWeights() {
		size_t n = 0;
		for (unsigned k = 0; k < getNumLayers(); k++)
			if (getLayer(k)->getType() == Layer::SUM_LAYER)
				for (unsigned l = 0; l < getLayer(k)->getNumNodes(); l++)
					n += getLayer(k)->getNode(l)->getNumChildren();
		return n;
	}

	size_t Network::getNumGaussians() {
		size_t n = 0;
		for (unsigned k = 0; k < getNumLayers(); k++)
			if (getLayer(k)->getType() == Layer::GAUSSIAN_LAYER)
				n += getLayer(k)->getNumNodes();
		return n;
	}

	size_t Network::getNumSums() {
		size_t n = 0;
		for (unsigned k = 0; k < getNumLayers(); k++)
			if (getLayer(k)->getType() == Layer::SUM_LAYER)
				n += getLayer(k)->getNumNodes();
		return n;
	}

	size_t Network::getNumProds() {
		size_t n = 0;
		for (unsigned k = 0; k < getNumLayers(); k++)
			if (getLayer(k)->getType() == Layer::PROD_LAYER)
				n += getLayer(k)->getNumNodes();
		return n;
	}

	void Network::performStructureCheck() {
		if (getNumLayers() == 0)
			throw std::runtime_error("error in Network::performStructureCheck(): Network has no layers.");
		if (layers_[0]->getType() != Layer::VAL_LAYER)
			throw std::runtime_error("error in Network::performStructureCheck(): layer 0 has to be of type Layer::VAL_LAYER.");

		unsigned k, l, m, n, o, p;
		Node *curNode, *childNode;

		// check feedforward
		structFeedforward_ = true;
		for (k = 0; k < getNumLayers() - 1; k++) {
			for (l = 0; l < getLayer(k)->getNumNodes(); l++) {
				curNode = getNode(k, l);
				for (m = k + 1; m < getNumLayers(); m++) {
					if (getLayer(m)->containsNode(curNode)) {
						structFeedforward_ = false;
						structComplete_ = false;
						structDecomposable_ = false;
						return;
					}
				}
			}
		}

		// check decomposability and completeness
		UMap<NodeId, Size_tVec > scopes;
		scopes.clear();
		Size_tVec curScope, tmpScope1, tmpScope2;
		Size_tVec::iterator it;

		structComplete_ = true;
		structDecomposable_ = true;

		for (k = 1; k < getNumLayers(); k++) {
			if ((!structComplete_) && (!structDecomposable_))
				break;
			for (l = 0; l < getLayer(k)->getNumNodes(); l++) {
				if ((!structComplete_) && (!structDecomposable_))
					break;

				curNode = getNode(k, l);
				if (structComplete_) {
					if ((curNode->getType() == Node::SUM_NODE) && (curNode->getNumChildren() > 1)) {
						for (m = 1; m < curNode->getNumChildren(); m++) {
							tmpScope1 = scopes[curNode->getChild(m - 1)->getId()];
							tmpScope2 = scopes[curNode->getChild(m)->getId()];

							if (tmpScope1.size() != tmpScope2.size()) {
								std::cout << "L " << k << " N " << l << std::endl;
								structComplete_ = false;
								break;
							}

							for (n = 0; n < tmpScope1.size(); n++) {
								if (tmpScope1[n] != tmpScope2[n]) {
									structComplete_ = false;
									break;
								}
							}

							if (!structComplete_)
								break;
						}
					}
				}

				if (structDecomposable_) {
					if ((curNode->getType() == Node::PROD_NODE) && (curNode->getNumChildren() > 1)) {
						for (m = 0; m < curNode->getNumChildren() - 1; m++) {
							tmpScope1 = scopes[curNode->getChild(m)->getId()];
							for (n = m + 1; n < curNode->getNumChildren(); n++) {
								tmpScope2 = scopes[curNode->getChild(n)->getId()];
								for (o = 0; o < tmpScope1.size(); o++) {
									for (p = 0; p < tmpScope2.size(); p++) {
										if (tmpScope1[o] == tmpScope2[p]) {
											structDecomposable_ = false;
											break;
										}
									}
									if (!structDecomposable_)
										break;
								}
								if (!structDecomposable_)
									break;
							}
							if (!structDecomposable_)
								break;
						}
					}
				}

				// make scope for current node
				curScope.clear();
				for (m = 0; m < curNode->getNumChildren(); m++) {
					childNode = curNode->getChild(m);
					if (childNode->getType() == Node::VAL_NODE) {
						curScope.push_back(getLayer(0)->getNodeIdx(childNode));
					}
					else {
						tmpScope1 = scopes[childNode->getId()];
						curScope.insert(curScope.end(), tmpScope1.begin(), tmpScope1.end());
					}
				}

				std::sort(curScope.begin(), curScope.end());
				it = std::unique(curScope.begin(), curScope.end());
				curScope.resize(it - curScope.begin());
				scopes[curNode->getId()] = curScope;
			}
		}
	}

	bool Network::isComplete() {
		return structComplete_;
	}

	bool Network::isDecomposable() {
		return structDecomposable_;
	}

	bool Network::isFeedforward() {
		return structFeedforward_;
	}

	Layer* Network::operator()(size_t layerIdx) {
		return getLayer(layerIdx);
	}

	Node* Network::operator()(size_t layerIdx, size_t idx) {
		return getNode(layerIdx, idx);
	}

	NetworkId Network::getId() {
		return id_;
	}

}
