/*
 *
 * The Network class implements feedforward sum-product networks.
 * A network is organized as a list of layers.
 *
 * When spn is a Network object, then spn(i) returns the i'th layer, and spn(i,j) returns the j'th node in the i'th layer.
 *
 * In a network spn, layer(0) is the bottom (input) layer of type ValLayer, and the highest layer (i.e. spn(spn.getNumLayers()-1))
 * contains the root node. Here it is assumed that the root is the first node in the highest layer, i.e. spn(spn.getNumLayers()-1, 0)
 *
 */

#ifndef NETWORK_HPP
#define	NETWORK_HPP

#include "TypeDefs.hpp"
#include "Node.hpp"
#include "Layer.hpp"
#include "GraphTools.hpp"
#include <queue>

namespace SPN {

	class Network;
	typedef std::vector<Network*> NetworkVec;
	typedef size_t NetworkId;

	class Network {
	private:
		NetworkId id_;
		static NetworkId idCounter_;

	protected:
		// list of layer pointers
		LayerVec layers_;

		// result flags of performStructureCheck()
		bool structComplete_;
		bool structDecomposable_;
		bool structFeedforward_;

	public:
		Network();
		virtual ~Network();

		// insert layer at a certain position
		virtual void insertLayer(Layer* layer, size_t idx);
		virtual void removeLayer(Layer* layer);
		virtual void addLayer(Layer* layer);

		// get methods
		virtual size_t getNumLayers();
		virtual LayerVec getLayers();
		virtual Layer* getLayer(size_t layerIdx);
		virtual Layer* getTopLayer();
		virtual size_t getLayerIdx(Layer* layer);
		virtual Node* getNode(size_t layerIdx, size_t idx);
		virtual bool containsLayer(Layer *layer);
		virtual size_t getLayerIdxOfNode(Node *node);

		// set values in the bottom layer
		virtual void setVals(const DoubleVec &values);
		virtual DoubleVec getVals();

		// eval the network from bottom to top
		virtual void eval();
		// backprop from top to bottom
		virtual void backprop();
		// 
		virtual double loglikelihood(DoubleVec &data);

		// MPE inference 
		virtual void MPEinference();

		// for all sum nodes in the network, switch between sum and max mode
		virtual void setSumMode();
		virtual void setMaxMode();

		// parameter initialization
		virtual void initWeightsRand();
		virtual void setGaussianMeansRand(double minVal, double maxVal);
		virtual void setGaussianSigmasRand(double minVal, double maxVal);
		virtual void setGaussianMeansUniform(double meanVal);
		virtual void setGaussianSigmasUniform(double sigmaVal);

		// remove all layers
		virtual void clear();

		// set weights to zero, which are smaller or equal than a certain threshold
		virtual void clipWeights(double threshold);
		virtual void removeSingletonParents(NodeVec &removedNodes, LayerVec &removedLayers);

		// disconnect 0-weighted children
		virtual void disconnectZeroChildren();

		// renormalize
		virtual void normalizeWeights();
		virtual void normalizeWeightsProbInvariant();

		// count number of weights, Gaussians, etc.
		virtual size_t getNumWeights();
		virtual size_t getNumGaussians();
		virtual size_t getNumSums();
		virtual size_t getNumProds();

		// for structure check: first call performStructureCheck(), 
		// then call isComplete(), isDecomposable(), or isFeedforward()
		virtual void performStructureCheck();
		virtual bool isComplete();
		virtual bool isDecomposable();
		virtual bool isFeedforward();

		Layer* operator()(size_t layerIdx);
		Node* operator()(size_t layerIdx, size_t idx);

		NetworkId getId();
	};

}


#endif	/* NETWORK_HPP */

