
#include "Node.hpp"

namespace SPN {

	NodeId Node::idCounter_ = 0;

	Node::Node(NodeType type) {
		if (
			(type != SUM_NODE) &&
			(type != PROD_NODE) &&
			(type != VAL_NODE) &&
			(type != INDICATOR_NODE) &&
			(type != GAUSSIAN_NODE))
			throw std::runtime_error("error in SPN::Node(int type): unknown type.");

		type_ = type;
		id_ = idCounter_++;
		children_.clear();
		parents_.clear();
		childInd_.clear();
		logVal_ = 0;
		logDerivative_ = -inf;
	}

	Node::~Node() {
		isolate();
	}

	NodeId Node::getId() const {
		return id_;
	}

	NodeType Node::getType() const {
		return type_;
	}

	void Node::forceLogVal(double logVal) {
		logVal_ = logVal;
	}

	double Node::getLogVal() const {
		return logVal_;
	}

	double Node::getLogDerivative() const {
		return logDerivative_;
	}

	double Node::getVal() const {
		return exp(logVal_);
	}

	double Node::getDerivative() const {
		return exp(logDerivative_);
	}

	void Node::backprop() {
		size_t numParents = getNumParents();
		if (numParents == 0) {
			logDerivative_ = 0.0;
			return;
		}

		double maxVal = -inf, derivative = 0.0;
		double *computeCache = new double[numParents];

		for (unsigned k = 0; k < numParents; k++) {
			Node *parent = getParent(k);
			computeCache[k] = parent->getLog_Dval_Dchild(this) + parent->getLogDerivative();
			if (computeCache[k] > maxVal)
				maxVal = computeCache[k];
		}

		if (!(maxVal > -inf)) {
			logDerivative_ = -inf;
			delete[] computeCache;
			return;
		}

		for (unsigned k = 0; k < numParents; k++)
			derivative += exp(computeCache[k] - maxVal);

		if (isnan(derivative))
			logDerivative_ = nan;
		else if (derivative > 0)
			logDerivative_ = maxVal + log(derivative);
		else
			logDerivative_ = -inf;

		delete[] computeCache;
	}

	void Node::connectChild(Node *child) {
		if ((isChild(child)) || (child->getId() == getId()))
			return;
		children_.push_back(child);
		child->parents_.push_back(this);
		childInd_.insert(child->getId());
	}

	void Node::connectParent(Node *parent) {
		if (isParent(parent) || (parent->getId() == getId()))
			return;
		parent->connectChild(this);
	}

	void Node::disconnectChild(Node *child) {
		NodeId childId = child->getId();

		bool found = false;
		for (unsigned k = 0; k < children_.size(); k++) {
			if (children_[k]->getId() == childId) {
				found = true;
				childInd_.erase(children_[k]->getId());
				children_.erase(children_.begin() + k);
				break;
			}
		}

		if (!found)
			throw std::runtime_error("Node::disconnectChild(Node &child): tried to disconnect non-child");

		found = false;
		for (unsigned k = 0; k < child->parents_.size(); k++) {
			if (child->parents_[k]->getId() == id_) {
				found = true;
				child->parents_.erase(child->parents_.begin() + k);
				break;
			}
		}

		if (!found)
			throw std::runtime_error("Node::disconnectChild(Node &child): inconsistent parent-child connection");
	}

	void Node::disconnectParent(Node *parent) {
		if (!isParent(parent))
			return;

		parent->disconnectChild(this);
	}

	void Node::isolate() {
		while (getNumParents() != 0)
			disconnectParent(getParent(getNumParents() - 1));

		while (getNumChildren() != 0)
			disconnectChild(getChild(getNumChildren() - 1));
	}

	NodeVec Node::getChildren() {
		return children_;
	}

	NodeVec Node::getParents() {
		return parents_;
	}

	Node* Node::getChild(size_t idx) {
		return (children_[idx]);
	}

	Node* Node::getParent(size_t idx) {
		return (parents_[idx]);
	}

	size_t Node::getNumChildren() {
		return children_.size();
	}

	size_t Node::getNumParents() {
		return parents_.size();
	}

	bool Node::isChild(Node *node) {
		return (childInd_.count(node->getId()) != 0);
	}

	bool Node::isParent(Node *node) {
		return node->isChild(this);
	}

	bool Node::operator==(const Node &rightNode) const {
		return (getId() == rightNode.getId());
	}
}
