/*
 *
 * The Node Class is an base class of the atomic units in SPNs: the nodes.
 * All node objects an unique id is assigned, currently implemented as unsigned
 * long.
 * We want to identify node objects uniquely in memory; Therefore, it is not
 * allowed to make copies of node objects, i.e. the copy constructor is private.
 * There is no public constructor, i.e. the Node class is actually abstract.
 * There is the protected Node(NodeType type) constructor which has to be called
 * from derived node classes.
 * There are currently 5 types of nodes:
 * sum nodes,
 * product nodes,
 * value nodes,
 * indicator nodes and
 * gaussian nodes.
 * The connectChild and connectParent allow you to connect nodes with each
 * other.
 * The eval() method represents the functionality of the node, and has to be
 * implemented in derived classes.
 * The backprop() method performs the backpropagation algorithm at *this node
 * object; this method is already implemented.
 * However, the backprop method needs the log of the derivative of each parent
 * nodes; therefore the getLog_Dval_Dchild(Node &child)
 * method has to be implemented in derived classes.
 *
 */


#ifndef NODE_HPP
#define	NODE_HPP

#include <vector>
#include "TypeDefs.hpp"
#include "Utils.hpp"


namespace SPN {

	class Node;
	typedef std::vector<Node*> NodeVec;
	typedef size_t NodeId;
	typedef int NodeType;


	///////////////////////
	/// Base class Node ///
	///////////////////////
	class Node {

	private:
		// this stores the type of the node (sum, product, ...)
		NodeType type_;

		// unique id of the node
		static NodeId idCounter_;
		NodeId id_;

		// connections are implemented as double-linked lists
		NodeVec children_;
		NodeVec parents_;

		// for quick check is some node is a child
		USet<NodeId> childInd_;

		Node(const Node& node);

	protected:
		Node(NodeType type);
		double logVal_;
		double logDerivative_;


	public:
		static const NodeType SUM_NODE = 0x01;
		static const NodeType PROD_NODE = 0x02;
		static const NodeType VAL_NODE = 0x03;
		static const NodeType INDICATOR_NODE = 0x04;
		static const NodeType GAUSSIAN_NODE = 0x05;

		virtual ~Node();

		virtual void eval() {
		}

		virtual double getLog_Dval_Dchild(Node *child) {
			throw std::runtime_error("double getLog_Dval_Dchild(Node *child): must not call");
			return 0.0;
		}

		virtual void backprop();

		// get value/derivative
		virtual double getLogVal() const;
		virtual double getLogDerivative() const;
		virtual double getVal() const;
		virtual double getDerivative() const;

		void forceLogVal(double logVal);

		// connect/disconnect
		virtual void connectChild(Node *child);
		virtual void connectParent(Node *parent);
		virtual void disconnectChild(Node *child);
		virtual void disconnectParent(Node *parent);

		// disconnect all parents/children
		void isolate();

		// number of children/parents
		size_t getNumChildren();
		size_t getNumParents();

		// get child or parent node with certain idx -- idx assigned in the order of connection
		Node* getChild(size_t idx);
		Node* getParent(size_t idx);

		// 
		NodeVec getChildren();
		NodeVec getParents();

		bool isChild(Node* node);
		bool isParent(Node* node);

		bool operator==(const Node &rigthNode) const;

		NodeId getId() const;
		int getType() const;
	};


	////////////////
	/// Sum Node ///
	////////////////
	class SumNode : public Node {
	private:
		UMap<NodeId, double> weights_;
		UMap<NodeId, double> logWeights_;

	protected:
		double maxLogVal_;
		unsigned maxIdx_;
		bool maxMode_;

	public:
		SumNode();
		virtual ~SumNode();

		virtual void eval();

		virtual double getLog_Dval_Dchild(Node *childId);
		virtual void connectChild(Node *child);
		virtual void connectChild(Node *child, double weight);
		virtual void disconnectChild(Node *child);

		virtual double getLogVal() const;
		virtual double getVal() const;

		virtual unsigned getMaxIdx() const;

		virtual void setWeight(Node *child, double weight);
		virtual void setWeights(const DoubleVec &weights);
		virtual void setLogWeight(Node *child, double logWeight);
		virtual void setLogWeights(const DoubleVec &logWeights);
		virtual double getWeight(Node *child);
		virtual double getLogWeight(Node *child);
		virtual void initWeightsRand();
		virtual void normalizeWeights();

		virtual void setSumMode();
		virtual void setMaxMode();
	};


	////////////////////
	/// Product Node ///
	////////////////////
	class ProdNode : public Node {
	protected:
		int zeroChildFlag_;

	public:
		ProdNode();
		virtual ~ProdNode();
		virtual void eval();
		virtual double getLogVal() const;
		virtual double getVal() const;
		virtual double getLog_Dval_Dchild(Node *child);
	};


	//////////////////
	/// Value Node ///
	//////////////////
	class ValNode : public Node {
	private:
		double val_;

	public:
		ValNode();
		ValNode(double logVal);
		virtual ~ValNode();
		virtual void eval();
		virtual double getLog_Dval_Dchild(Node *child);

		virtual double getLogVal() const;
		virtual double getVal() const;
		virtual void setLogVal(double logVal);
		virtual void setVal(double val);
	};


	//////////////////////
	/// Indicator Node ///
	//////////////////////
	class IndicatorNode : public Node {
	protected:
		UMap<NodeId, double> indValues_;
		int clampVal_;
		static double tol_;

	public:
		IndicatorNode();
		virtual ~IndicatorNode();
		virtual void eval();
		virtual double getLog_Dval_Dchild(Node *child);

		virtual void connectChild(Node *child);
		virtual void connectChild(Node *child, double indValue);
		virtual void disconnectChild(Node *child);

		virtual double getLogVal() const;
		virtual double getVal() const;

		virtual void setIndValue(Node *child, double indValue);
		virtual double getIndValue(Node *child);

		virtual void clamp(bool clamVal);
		virtual void unclamp();

		static void setTolerance(double tol);
		static double getTolerance();
	};


	/////////////////////
	/// Gaussian Node ///
	/////////////////////
	class GaussianNode : public Node {
	protected:
		UMap<NodeId, double> means_;
		UMap<NodeId, double> sigmas_;		
		int mode_;

	public:
		static const double logSqrt2Pi;
		static const int PDF_MODE = 1;
		static const int CDF_MODE = 2;

		GaussianNode();
		virtual ~GaussianNode();
		virtual void eval();
		virtual double getLog_Dval_Dchild(Node *child);

		virtual void connectChild(Node *child);
		virtual void connectChild(Node *child, double mean, double sigma);
		virtual void disconnectChild(Node *child);

		virtual void setMean(Node *child, double mean);
		virtual double getMean(Node *child);
		virtual void setMeansRand(double minVal, double maxVal);
		virtual void setSigma(Node *child, double sigma);
		virtual double getSigma(Node *child);
		virtual void setSigmasRand(double minVal, double maxVal);
		
		virtual void setMode(int mode);
	};


}

#endif	/* NODE_HPP */

