
#include "PDLearning.hpp"
#include <iostream>

namespace SPN {

	PDLearning::PDLearning() {
		pixelOrder_ = ROW_MAJOR;
		numChannels_ = GRAY;
		width_ = 64;
		height_ = 64;
		numGaussPerPixel_ = 4;
		maxNumSumsPerRectangle_ = 20;
		coarseRes_.clear();
		coarseResX_.clear();
		coarseResY_.clear();
		gaussMeans_.clear();

		options.validationDataSource = NULL;
		options.maxIter = 30;
		options.sparsePrior = 1.0;
		options.sparsePriorAnneal = 10;
		options.smoothCount = 0.01;
		options.incrEM_batchSize = 50;
		options.stopping_objMinAbsChange = 0.1;
		// the relative stopping criterion was not considerd in the code of PD.
		// Therefore we set it to 10^-21 as default
		options.stopping_objMinRelChange = -1e21;
		options.earlyStoppingKrel = 3;
		options.earlyStoppingKabs = 5;

		preSubtractCounts_ = true;
		flatEM_ = true;
	}

	PDLearning::~PDLearning() {
	}

	void PDLearning::setRowMajor() {
		pixelOrder_ = ROW_MAJOR;
	}

	void PDLearning::setColumnMajor() {
		pixelOrder_ = COLUMN_MAJOR;
	}

	void PDLearning::setNumChannels(int numChannels) {
		if (numChannels >= 1)
			numChannels_ = numChannels;
		else
			numChannels_ = 1;
	}

	void PDLearning::setMaxNumSumsPerRectangle(int maxNumSums) {
		if (maxNumSums >= 1)
			maxNumSumsPerRectangle_ = maxNumSums;
		else
			maxNumSumsPerRectangle_ = 1;
	}

	void PDLearning::setNumGaussPerPixel(int numGauss) {
		if (numGauss >= 1)
			numGaussPerPixel_ = numGauss;
		else
			numGaussPerPixel_ = 1;
	}

	void PDLearning::setCoarseRes(unsigned coarseRes) {
		setCoarseRes(UnsignedVec(1, coarseRes));
	}

	void PDLearning::setCoarseRes(UnsignedVec coarseRes) {
		UnsignedVec::iterator it;
		coarseRes_ = coarseRes;

		for (unsigned k = 0; k < coarseRes_.size(); k++) {
			if (coarseRes_[k] <= 0)
				coarseRes_[k] = 1;
		}

		std::sort(coarseRes_.begin(), coarseRes_.end());
		it = std::unique(coarseRes_.begin(), coarseRes_.end());
		coarseRes_.resize(it - coarseRes_.begin());

		if (coarseRes_.empty()) {
			coarseRes_.resize(1);
			coarseRes_[0] = 1;
		}
	}

	void PDLearning::setCoarseResX(unsigned coarseResX) {
		setCoarseResX(UnsignedVec(1,coarseResX));
	}

	void PDLearning::setCoarseResX(UnsignedVec coarseResX) {
		UnsignedVec::iterator it;
		coarseResX_ = coarseResX;

		for (unsigned k = 0; k < coarseResX_.size(); k++) {
			if (coarseResX_[k] <= 0)
				coarseResX_[k] = 1;
		}

		std::sort(coarseResX_.begin(), coarseResX_.end());
		it = std::unique(coarseResX_.begin(), coarseResX_.end());
		coarseResX_.resize(it - coarseResX_.begin());
	}

	void PDLearning::setCoarseResY(unsigned coarseResY) {
		setCoarseResY(UnsignedVec(1, coarseResY));
	}

	void PDLearning::setCoarseResY(UnsignedVec coarseResY) {
		UnsignedVec::iterator it;
		coarseResY_ = coarseResY;

		for (unsigned k = 0; k < coarseResY_.size(); k++) {
			if (coarseResY_[k] <= 0)
				coarseResY_[k] = 1;
		}

		std::sort(coarseResY_.begin(), coarseResY_.end());
		it = std::unique(coarseResY_.begin(), coarseResY_.end());
		coarseResY_.resize(it - coarseResY_.begin());
	}

	void PDLearning::setDims(int width, int height) {
		if (width >= 1)
			width_ = width;
		else
			width_ = 1;

		if (height >= 1)
			height_ = height;
		else
			height_ = 1;
	}

	void PDLearning::setGaussMeans(DoubleVec2 means) {
		gaussMeans_ = means;
	}

	unsigned PDLearning::getInputIdx(int x, int y, int c) {
		unsigned idx;

		if (pixelOrder_ == ROW_MAJOR)
			idx = (unsigned)(y * numChannels_ * width_ + c * width_ + x);
		else
			idx = (unsigned)(x * numChannels_ * height_ + c * height_ + y);

		return idx;
	}

	bool PDLearning::incCounter(std::vector<unsigned> &counter, std::vector<unsigned> &numVals) {
		bool ret = false;
		for (unsigned k = 0; k < counter.size(); k++) {
			if (counter[k] < numVals[k] - 1) {
				counter[k]++;
				return ret;
			}
			else {
				counter[k] = 0;
			}
		}
		ret = true;
		return ret;
	}

	void PDLearning::assertRectangles(PDRectangle *parent, PDRectangle *child1, PDRectangle *child2) {
		bool assertFail = false;

		if (!((parent->width_ == child1->width_) && (parent->width_ == child2->width_)) &&
			!((parent->height_ == child1->height_) && (parent->height_ == child2->height_)))
			assertFail = true;
		else {
			if ((parent->width_ == child1->width_) && (parent->width_ == child2->width_)) {
				if (parent->height_ != child1->height_ + child2->height_)
					assertFail = true;

				if (child1->y_ <= child2->y_) {
					if ((parent->y_ != child1->y_) || (child2->y_ != child1->y_ + child1->height_))
						assertFail = true;
				}
				else {
					if ((parent->y_ != child2->y_) || (child1->y_ != child2->y_ + child2->height_))
						assertFail = true;
				}
			}
			else {
				if (parent->width_ != child1->width_ + child2->width_)
					assertFail = true;

				if (child1->x_ <= child2->x_) {
					if ((parent->x_ != child1->x_) || (child2->x_ != child1->x_ + child1->width_))
						assertFail = true;
				}
				else {
					if ((parent->x_ != child2->x_) || (child1->x_ != child2->x_ + child2->width_))
						assertFail = true;
				}
			}
		}

		if (assertFail) {
			std::cout << parent->x_ << " " << parent->width_ << "  " << parent->y_ << " " << parent->height_ << std::endl;
			std::cout << child1->x_ << " " << child1->width_ << "  " << child1->y_ << " " << child1->height_ << std::endl;
			std::cout << child2->x_ << " " << child2->width_ << "  " << child2->y_ << " " << child2->height_ << std::endl;
			throw std::runtime_error("rectangle assertion failed");
		}
	}

	void PDLearning::setWeights(std::vector<PDRectangle* > &rectangles, const StatsCounterFloat &counts) {

		for (unsigned k = 0; k < rectangles.size(); k++) {
			PDRectangle *curRectangle = rectangles[k];

			for (std::map<unsigned, Node* >::iterator it = curRectangle->nodePtrs_.begin();
				it != curRectangle->nodePtrs_.end();
				it++) {
				if (it->second->getType() != Node::SUM_NODE)
					continue;

				SumNode* curSumNode = (SumNode*)it->second;
				double totalCount = counts.get(*curSumNode);

				if (totalCount <= numericZero)
					for (unsigned pC = 0; pC < curSumNode->getNumChildren(); pC++)
						curSumNode->setWeight(curSumNode->getChild(pC), 0.0);
				else {
					DoubleVec tmpCounts;
					tmpCounts.clear();
					std::vector<ProdNode*> nonZeroChildren;
					nonZeroChildren.clear();

					for (unsigned pC = 0; pC < curSumNode->getNumChildren(); pC++) {
						ProdNode *curProdNode = (ProdNode*)(curSumNode->getChild(pC));
						double count = counts.get(*curSumNode, *curProdNode);
						if (count > numericZero) {
							tmpCounts.push_back(count + options.smoothCount);
							nonZeroChildren.push_back(curProdNode);
						}
						else
							curSumNode->setWeight(curSumNode->getChild(pC), 0.0);
					}

					if (nonZeroChildren.size() != counts.getNumNonzeroChildren(*curSumNode))
						throw std::logic_error("# of nonzero children inconsistent");

					double totalSmooth = totalCount + nonZeroChildren.size() * options.smoothCount;
					for (unsigned pC = 0; pC < nonZeroChildren.size(); pC++)
						curSumNode->setWeight(nonZeroChildren[pC], tmpCounts[pC] / totalSmooth);
				}
			}
		}
	}

	Network* PDLearning::learn(DataSource* trainData, NetworkFactory &factory) {

		unsigned numSums = 0;
		unsigned numProds = 0;
		int numLevels = 0;

		std::vector<PDRectangle* > rectangles;
		std::vector<PDPartition* > partitions;
		std::map<PDRectangle, PDRectangle* > rectangleMap;
		std::map<unsigned, std::vector<PDRectangle*> > rectanglesOfLevel;

		rectangles.clear();
		partitions.clear();
		rectangleMap.clear();

		///////////////////////////////////////////////////////////////////
		// Determine Structure                                           //
		// we start with a single rectangle and decompose it further     //
		///////////////////////////////////////////////////////////////////
		PDRectangle* root = new PDRectangle;
		root->x_ = 0;
		root->y_ = 0;
		root->width_ = width_;
		root->height_ = height_;

		// rectangles will contain a list of pointers to all generated rectangles
		rectangles.push_back(root);
		rectangleMap[*root] = root;

		for (unsigned k = 0; k < rectangles.size(); k++) {
			// if rectangle was already split, or if it is of unit size, don't split
			if ((rectangles[k]->childPartitions_.empty() == false)
				|| ((rectangles[k]->width_ == 1) && (rectangles[k]->height_ == 1)))
				continue;

			PDRectangle *curRectangle = rectangles[k];

			//
			// deltaX, deltaY denote the used resolutions 
			//
			unsigned deltaX = 1;
			unsigned deltaY = 1;

			if ((coarseResX_.empty()) || (coarseResY_.empty())) {
				for (unsigned l = 0; l < coarseRes_.size(); l++) {
					size_t idx = coarseRes_.size() - 1 - l;
					if (
						(coarseRes_[idx] <= curRectangle->width_) &&
						(coarseRes_[idx] <= curRectangle->height_) &&
						((coarseRes_[idx] != curRectangle->height_) ||
						(coarseRes_[idx] != curRectangle->width_))) {
						deltaX = coarseRes_[idx];
						deltaY = coarseRes_[idx];
						break;
					}
				}
			}
			else {
				for (unsigned l = 0; l < coarseResX_.size(); l++) {
					size_t idx = coarseResX_.size() - 1 - l;
					if (coarseResX_[idx] < curRectangle->width_) {
						deltaX = coarseResX_[idx];
						break;
					}
				}
				for (unsigned l = 0; l < coarseResY_.size(); l++) {
					size_t idx = coarseResY_.size() - 1 - l;
					if (coarseResY_[idx] < curRectangle->height_) {
						deltaY = coarseResY_[idx];
						break;
					}
				}
			}

			// split into smaller rectangles -- x dimension
			PDRectangle r1, r2;
			PDRectangle *r1ptr;
			PDRectangle *r2ptr;

			r1.y_ = curRectangle->y_;
			r1.height_ = curRectangle->height_;
			r1.x_ = curRectangle->x_;
			r2.y_ = curRectangle->y_;
			r2.height_ = curRectangle->height_;

			for (unsigned x = deltaX; x < curRectangle->width_; x = x + deltaX) {
				r1.width_ = x;
				r2.x_ = curRectangle->x_ + x;
				r2.width_ = curRectangle->width_ - x;

				if (rectangleMap.count(r1) > 0)
					r1ptr = rectangleMap[r1];
				else {
					r1ptr = new PDRectangle;
					*r1ptr = r1;
					rectangleMap[r1] = r1ptr;
					rectangles.push_back(r1ptr);
				}
				if (rectangleMap.count(r2) > 0)
					r2ptr = rectangleMap[r2];
				else {
					r2ptr = new PDRectangle;
					*r2ptr = r2;
					rectangleMap[r2] = r2ptr;
					rectangles.push_back(r2ptr);
				}

				PDPartition *p = new PDPartition;
				partitions.push_back(p);
				p->parentRectangle_ = rectangles[k];
				p->childRectangles_.push_back(r1ptr);
				p->childRectangles_.push_back(r2ptr);
				curRectangle->childPartitions_.push_back(p);
				r1ptr->parentPartitions_.push_back(p);
				r2ptr->parentPartitions_.push_back(p);
			}

			// split into smaller rectangles -- y dimension
			r1.y_ = curRectangle->y_;
			r1.width_ = curRectangle->width_;
			r1.x_ = curRectangle->x_;
			r2.x_ = curRectangle->x_;
			r2.width_ = curRectangle->width_;

			for (unsigned y = deltaY; y < curRectangle->height_; y = y + deltaY) {
				r1.height_ = y;
				r2.y_ = curRectangle->y_ + y;
				r2.height_ = curRectangle->height_ - y;

				if (rectangleMap.count(r1) > 0)
					r1ptr = rectangleMap[r1];
				else {
					r1ptr = new PDRectangle;
					*r1ptr = r1;
					rectangleMap[r1] = r1ptr;
					rectangles.push_back(r1ptr);
				}
				if (rectangleMap.count(r2) > 0)
					r2ptr = rectangleMap[r2];
				else {
					r2ptr = new PDRectangle;
					*r2ptr = r2;
					rectangleMap[r2] = r2ptr;
					rectangles.push_back(r2ptr);
				}

				PDPartition *p = new PDPartition;
				partitions.push_back(p);
				p->parentRectangle_ = rectangles[k];
				p->childRectangles_.push_back(r1ptr);
				p->childRectangles_.push_back(r2ptr);
				curRectangle->childPartitions_.push_back(p);
				r1ptr->parentPartitions_.push_back(p);
				r2ptr->parentPartitions_.push_back(p);
			}
		}

		//
		// set levels
		//
		root->setLevelRecursive(0);
		numLevels = -1;
		for (unsigned k = 0; k < rectangles.size(); k++)
			if ((rectangles[k]->level_ + 1) > numLevels)
				numLevels = rectangles[k]->level_ + 1;

		std::cout << "#rectangle levels " << numLevels << std::endl;

		//
		// set all rectangles of size 1x1 (pixels) to the highest level
		//
		for (unsigned k = 0; k < rectangles.size(); k++)
			if ((rectangles[k]->width_ == 1) && (rectangles[k]->height_ == 1))
				rectangles[k]->level_ = numLevels - 1;

		//
		// organize level map of rectangles
		//
		for (unsigned k = 0; k < rectangles.size(); k++)
			rectanglesOfLevel[rectangles[k]->level_].push_back(rectangles[k]);

		std::cout << "#rectangles " << rectangles.size() << std::endl;

		/////////////////////////
		// Parameter Learning  //
		/////////////////////////

		// we require that there are at least as many samples as Gaussians...
		if (trainData->getNumSamples() < (size_t) numGaussPerPixel_)
			throw std::runtime_error("PDLearning::makeNetwork(Network &network): "
			"data source has less samples than desired number of Gaussians.");

		// generate input layer
		ValLayer *inputLayer = (ValLayer*)factory.generateLayer(Layer::VAL_LAYER,
			width_ * height_ * numChannels_);

		// generate Gauss layer
		GaussianLayer *gaussLayer = (GaussianLayer*)factory.generateLayer(Layer::GAUSSIAN_LAYER, 0);

		//
		// some checks
		//
		if (!gaussMeans_.empty()) {
			if (!isMat(gaussMeans_))
				throw std::runtime_error("bad pre-defined gauss means (not a matrix)");
			if (gaussMeans_.size() != (unsigned)numGaussPerPixel_)
				throw std::runtime_error("bad pre-defined gauss means");
			if (gaussMeans_[0].size() != (unsigned)width_ * height_ * numChannels_)
				throw std::runtime_error("bad pre-defined gauss means");
		}

		//
		// Gaussian are initialized using the means of 25% quantiles
		//
		double numSamplesPerGauss =
			(double)(trainData->getNumSamples()) / (double)(numGaussPerPixel_);
		for (unsigned k = 0; k < rectanglesOfLevel[numLevels - 1].size(); k++) {
			unsigned pixX = rectanglesOfLevel[numLevels - 1][k]->x_;
			unsigned pixY = rectanglesOfLevel[numLevels - 1][k]->y_;
			for (int c = 0; c < numChannels_; c++) {
				unsigned inputIdx = getInputIdx(pixX, pixY, c);

				if (gaussMeans_.empty()) {
					DoubleVec tmpData;
					tmpData.clear();
					for (size_t dC = 0; dC < trainData->getNumSamples(); dC++)
						tmpData.push_back(trainData->getSample(dC)[inputIdx]);

					std::sort(tmpData.begin(), tmpData.end());

					unsigned quantileStart = 0;
					for (int g = 0; g < numGaussPerPixel_; g++) {
						GaussianNode *gaussNode = (GaussianNode*)factory.generateNode(Node::GAUSSIAN_NODE);
						gaussNode->connectChild((*inputLayer)(inputIdx));
						rectanglesOfLevel[numLevels - 1][k]->nodePtrs_[g] = gaussNode;

						unsigned quantileEnd = (unsigned)(round((g + 1) * numSamplesPerGauss));
						double quantileMean = 0.0;
						for (unsigned sC = quantileStart; sC < quantileEnd; sC++)
							quantileMean += tmpData[sC];
						quantileMean = quantileMean / (double)(quantileEnd - quantileStart);
						quantileStart = quantileEnd;
						gaussNode->setMean((*inputLayer)(inputIdx), quantileMean);
						gaussNode->setSigma((*inputLayer)(inputIdx), 1.0);
						gaussLayer->insertNode(gaussNode);
					}
				}
				else {
					for (int g = 0; g < numGaussPerPixel_; g++) {
						GaussianNode *gaussNode = (GaussianNode*)factory.generateNode(Node::GAUSSIAN_NODE);
						gaussNode->connectChild((*inputLayer)(inputIdx));
						rectanglesOfLevel[numLevels - 1][k]->nodePtrs_[g] = gaussNode;

						gaussNode->setMean((*inputLayer)(inputIdx), gaussMeans_[g][inputIdx]);
						gaussNode->setSigma((*inputLayer)(inputIdx), 1);
						gaussLayer->insertNode(gaussNode);
					}
				}
			}
		}

		/////////////////////////////////////////////////
		/// Generate Sum nodes for all larger regions ///
		/////////////////////////////////////////////////

		// maps sumNodes to rectangles
		std::map<NodeId, PDRectangle*> sumNode_rectanglePtr;
		
		//
		// counter
		//
		std::vector<StatsCounterFloat> batchCount;
		StatsCounterFloat oldBatchCount;
		StatsCounterFloat totalCount;

		// maps prodNodes (via their node id) to PDid and vice versa
		std::map<NodeId, PDProdNodeId> prodNode_PDid;
		std::map<PDProdNodeId, Node*> PDid_prodNode;
		
		double logLikelihood;
		double oldLogLikelihood = 0.0;
		double valLogLikelihood;
		double oldValLogLikelihood = 0.0;
		double bestValLL = -inf;
		Network* bestValSPN = NULL;
		unsigned numValLLnotInc = 0;
		unsigned numBestValLLnotInc = 0;

		size_t numBatches, batchSize;
		double sparsePrior;

		// determine batch size
		if ((options.incrEM_batchSize > 0) && ((size_t) options.incrEM_batchSize < trainData->getNumSamples())) {
			// incremental EM
			if ((trainData->getNumSamples() % (unsigned)options.incrEM_batchSize) == 0)
				numBatches = trainData->getNumSamples() / (unsigned)options.incrEM_batchSize;
			else
				numBatches = (trainData->getNumSamples() / (unsigned)options.incrEM_batchSize) + 1;
			batchSize = (unsigned)options.incrEM_batchSize;
		}
		else {
			// standard EM (not incremental)
			numBatches = 1;
			batchSize = trainData->getNumSamples();
		}

		batchCount.resize(numBatches);

		// start training 
		for (unsigned iterC = 0; iterC < options.maxIter; iterC++) {

			// anneal the sparse prior
			if (iterC < options.sparsePriorAnneal)
				sparsePrior = (((double)(iterC + 1)) / ((double)options.sparsePriorAnneal)) * options.sparsePrior;
			else
				sparsePrior = options.sparsePrior;

			std::cout << "iter " << iterC << ", sparse prior " << sparsePrior << std::endl;

			//
			// iteration over batches
			//
			unsigned sampleCounter = 0;
			unsigned progressCounter = 0;

			for (unsigned bC = 0; bC < numBatches; bC++) {
				std::cout << "." << std::flush;
				if (progressCounter++ == 79) {
					std::cout << std::endl;
					progressCounter = 0;
				}

				oldBatchCount = batchCount[bC];
				batchCount[bC].clear();

				if (preSubtractCounts_) {
					totalCount -= oldBatchCount;
					setWeights(rectangles, totalCount);
				}

				for (unsigned batchM = 0;
					((sampleCounter < trainData->getNumSamples()) && (batchM < batchSize));
					batchM++, sampleCounter++) {

					//
					// map from sumNodeId -> MAP product node of this sum node
					//
					std::map<NodeId, PDProdNodeId> mapProdNodes;
					mapProdNodes.clear();

					inputLayer->setVals(trainData->getSample(sampleCounter));

					/////////////////////
					/// Upwards pass  ///
					/// bottom to top ///
					/////////////////////
					for (unsigned levelCount = 0; levelCount < (unsigned)numLevels; levelCount++) {
						unsigned curLevelIdx = numLevels - 1 - levelCount;

						//
						// Gaussian layer	  
						//
						if (curLevelIdx == (unsigned)numLevels - 1) {
							for (unsigned l = 0; l < rectanglesOfLevel[curLevelIdx].size(); l++) {
								double maxVal = -inf;
								unsigned maxIdx = 0;

								for (unsigned n = 0; n < rectanglesOfLevel[curLevelIdx][l]->nodePtrs_.size(); n++) {
									rectanglesOfLevel[curLevelIdx][l]->nodePtrs_[n]->eval();
									double curVal = rectanglesOfLevel[curLevelIdx][l]->nodePtrs_[n]->getLogVal();

									// P&D do not normalize by sqrt(2*pi)
									// this relevant due to sparse prior
									curVal += GaussianNode::logSqrt2Pi;
									rectanglesOfLevel[curLevelIdx][l]->nodePtrs_[n]->forceLogVal(curVal);
									if ((n == 0) || (curVal > maxVal)) {
										maxVal = curVal;
										maxIdx = n;
									}
								}

								rectanglesOfLevel[curLevelIdx][l]->MAPNodeIdx_ = maxIdx;
								rectanglesOfLevel[curLevelIdx][l]->MAPNodeLogVal_ = maxVal;
							}
							continue;
						}

						//
						// Sum and Product layers
						//
						for (unsigned l = 0; l < rectanglesOfLevel[curLevelIdx].size(); l++) {
							PDRectangle *curRectangle = rectanglesOfLevel[curLevelIdx][l];

							// Iterate over all partitions, i.e. over all possibilities to 
							// decompose the current rectangle into two smaller ones.
							// For each partition, find the MAP product node, which may not yet
							// be physically existing
							std::vector<PDProdNodeId> prodNodeOptions;
							double maxVal = -inf;

							for (unsigned n = 0; n < curRectangle->childPartitions_.size(); n++) {
								PDRectangle *r1ptr = curRectangle->childPartitions_[n]->childRectangles_[0];
								PDRectangle *r2ptr = curRectangle->childPartitions_[n]->childRectangles_[1];

								double curVal = r1ptr->MAPNodeLogVal_ + r2ptr->MAPNodeLogVal_;

								if ((n == 0) || (curVal > maxVal)) {
									maxVal = curVal;
									prodNodeOptions.clear();
								}
								if (curVal == maxVal)
									prodNodeOptions.push_back(PDProdNodeId(
									*r1ptr, *r2ptr,
									r1ptr->MAPNodeIdx_, r2ptr->MAPNodeIdx_));
							}

							// break ties
							PDProdNodeId defMapProdNodeId = prodNodeOptions[rand() % prodNodeOptions.size()];
							double defMapProdNodeVal = maxVal;

							//
							// evaluate the EXISTING product nodes in this rectangle 
							//
							for (unsigned n = 0; n < curRectangle->prodNodePtrs_.size(); n++)
								curRectangle->prodNodePtrs_[n]->eval();

							// evaluate the sum nodes in this rectangle, and find their MAP children
							std::vector<unsigned> sumNodeOptions;
							sumNodeOptions.clear();
							double maxSumNodeVal = -inf;
							bool consideredUnusedNode = false;

							// for (unsigned curSumIdx = 0;
							//         curSumIdx < curRectangle->nodePtrs_.size();
							//         curSumIdx++) {
							for (std::map<unsigned, Node* >::iterator it = curRectangle->nodePtrs_.begin();
								it != curRectangle->nodePtrs_.end();
								it++) {

								unsigned curSumIdx = it->first;
								SumNode* curSumNode = (SumNode*)it->second;
								//SumNode* curSumNode = (SumNode*) curRectangle->nodePtrs_[curSumIdx];
								curSumNode->eval();

								// is this an unused node (i.e. it has no children)?
								// all unused nodes are equal, so we have to consider only one
								if (totalCount.getNumNonzeroChildren(*curSumNode) == 0) {
									if (consideredUnusedNode)
										continue;
									else
										consideredUnusedNode = true;
								}

								double normConst = totalCount.get(*curSumNode) +
									totalCount.getNumNonzeroChildren(*curSumNode) * options.smoothCount;
								double denormVal = curSumNode->getLogVal() + log(normConst);

								if (isnan(denormVal))
									throw std::logic_error("denormVal == nan");

								// MAP of existing children ...
								bool defMAPProdIsChild = false;
								double maxVal = -inf;
								for (unsigned pC = 0; pC < curSumNode->getNumChildren(); pC++) {
									Node* curChild = curSumNode->getChild(pC);

									if (totalCount.get(*curSumNode, *curChild) < numericZero)
										continue;

									// flatEM_: original PD code
									double curVal = flatEM_ ?
										logSumExp(denormVal, curSumNode->getChild(pC)->getLogVal())
										:
										curSumNode->getChild(pC)->getLogVal()
										+ log(totalCount.get(*curSumNode, *curChild) + 1);

									if (curVal > maxVal) {
										maxVal = curVal;
										prodNodeOptions.clear();
									}
									if (curVal == maxVal)
										prodNodeOptions.push_back(prodNode_PDid[curChild->getId()]);

									if (prodNode_PDid[curChild->getId()] == defMapProdNodeId)
										defMAPProdIsChild = true;
								}

								// ... default MAP child, if not child of the sum node
								if (!defMAPProdIsChild) {
									double curVal = flatEM_ ?
										logSumExp(denormVal, defMapProdNodeVal) - sparsePrior
										:
										defMapProdNodeVal - sparsePrior;

									if ((curVal > maxVal) || (totalCount.getNumNonzeroChildren(*curSumNode) == 0)) {
										maxVal = curVal;
										prodNodeOptions.clear();
										prodNodeOptions.push_back(defMapProdNodeId);
									}
								}

								// break ties among product nodes
								mapProdNodes[curSumNode->getId()] = prodNodeOptions[rand() % prodNodeOptions.size()];

								// new value for the sum node
								double curVal = 0.0;
								if (flatEM_)
									curVal = maxVal - log(normConst + 1.0);
								else {
									Node* prodNode = 0;
									if (PDid_prodNode.count(mapProdNodes[curSumNode->getId()]) != 0)
										prodNode = PDid_prodNode[mapProdNodes[curSumNode->getId()]];

									if ((prodNode == 0) || (totalCount.get(*curSumNode, *prodNode) <= numericZero))
										curVal =
										logSumExp(denormVal, defMapProdNodeVal) - sparsePrior
										- log(normConst + 1.0);
									else
										curVal =
										logSumExp(denormVal, prodNode->getLogVal()) - log(normConst + 1.0);
								}
								curSumNode->forceLogVal(curVal);
								
								if ((curSumIdx == 0) || (curVal > maxSumNodeVal)) {
									sumNodeOptions.clear();
									maxSumNodeVal = curVal;
								}
								if (curVal == maxSumNodeVal)
									sumNodeOptions.push_back(curSumIdx);
							}

							// the preceding loop was over physically existing sum nodes;
							// possibly, we have to consider a not yet existing one
							if (!consideredUnusedNode) {
								bool foundEmpty = false;
								unsigned emptyIdx = 0;
								unsigned maxSumNodes = (curLevelIdx == 0) ? 1 : maxNumSumsPerRectangle_;
								for (unsigned sk = 0; sk < maxSumNodes; sk++) {
									if (curRectangle->nodePtrs_.count(sk) == 0) {
										foundEmpty = true;
										emptyIdx = sk;
										break;
									}
								}

								if (foundEmpty) {
									double curVal = defMapProdNodeVal - sparsePrior;
									if ((curVal > maxSumNodeVal) || (sumNodeOptions.empty())) {
										sumNodeOptions.clear();
										maxSumNodeVal = curVal;
										sumNodeOptions.push_back(emptyIdx);
									}
								}
							}

							// break ties among sum nodes
							curRectangle->MAPNodeIdx_ = sumNodeOptions[rand() % sumNodeOptions.size()];
							curRectangle->MAPNodeLogVal_ = maxSumNodeVal;
							// if a non-existing sum node was selected, generate it now
							if (curRectangle->nodePtrs_.count(curRectangle->MAPNodeIdx_) == 0) {
								SumNode* newSumNode = (SumNode*)factory.generateNode(Node::SUM_NODE);
								numSums++;
								curRectangle->nodePtrs_[curRectangle->MAPNodeIdx_] = newSumNode;
								sumNode_rectanglePtr[newSumNode->getId()] = curRectangle;
								mapProdNodes[newSumNode->getId()] = defMapProdNodeId;
								newSumNode->forceLogVal(defMapProdNodeVal - sparsePrior);
							}
						}
					}
					// end upward pass

					// down pass
					if (rectanglesOfLevel[0][0]->nodePtrs_.count(0) == 0)
						throw std::runtime_error("root sum node not generated");

					std::vector<SumNode*> parseStack;
					parseStack.clear();
					parseStack.push_back((SumNode*)rectanglesOfLevel[0][0]->nodePtrs_[0]);

					for (unsigned k = 0; k < parseStack.size(); k++) {
						SumNode* curSumNode = parseStack[k];
						PDProdNodeId curMapProdNodeId = mapProdNodes[curSumNode->getId()];

						// if prod node does not exist, create it now
						ProdNode* curProdNode = 0;
						if (PDid_prodNode.count(curMapProdNodeId) == 0) {
							curProdNode = (ProdNode*)factory.generateNode(Node::PROD_NODE);
							numProds++;
							// establish mapping
							PDid_prodNode[curMapProdNodeId] = curProdNode;
							prodNode_PDid[curProdNode->getId()] = curMapProdNodeId;
							// register product node in rectangle
							sumNode_rectanglePtr[curSumNode->getId()]->prodNodePtrs_.push_back(curProdNode);
						}
						else
							curProdNode = (ProdNode*)PDid_prodNode[curMapProdNodeId];

						PDRectangle *r1ptr = rectangleMap[curMapProdNodeId.getRectangle1()];
						PDRectangle *r2ptr = rectangleMap[curMapProdNodeId.getRectangle2()];

						if (r1ptr->nodePtrs_.count(curMapProdNodeId.getIdx1()) == 0)
							throw std::runtime_error("sum node in parse tree not generated");
						if (r2ptr->nodePtrs_.count(curMapProdNodeId.getIdx2()) == 0)
							throw std::runtime_error("sum node in parse tree not generated");

						Node* childNode1 = r1ptr->nodePtrs_[curMapProdNodeId.getIdx1()];
						Node* childNode2 = r2ptr->nodePtrs_[curMapProdNodeId.getIdx2()];

						// connect...
						if (!curSumNode->isChild(curProdNode)) {
							curSumNode->connectChild(curProdNode);
							batchCount[bC].set(*curSumNode, *curProdNode, 0.0);
						}
						if (!curProdNode->isChild(childNode1))
							curProdNode->connectChild(childNode1);
						if (!curProdNode->isChild(childNode2))
							curProdNode->connectChild(childNode2);

						// update counts
						batchCount[bC].add(*curSumNode, *curProdNode, 1.0);

						if (r1ptr->level_ < numLevels - 1)
							parseStack.push_back((SumNode*)childNode1);
						if (r2ptr->level_ < numLevels - 1)
							parseStack.push_back((SumNode*)childNode2);
					}
				} // sampleCount

				//std::cout << "entries in batch map: " << batchCount[bC].getNumEntries() << std::endl;
				//std::cout << "numSums: " << numSums << std::endl;
				//std::cout << "numProds: " << numProds << std::endl;

				// update the statistics of this batch 
				if (!preSubtractCounts_)
					totalCount -= oldBatchCount;

				totalCount += batchCount[bC];

				// set weights
				if (!preSubtractCounts_)
					setWeights(rectangles, totalCount);
			} // end for batchCount

			if (preSubtractCounts_)
				setWeights(rectangles, totalCount);

			// logLikelihood
			logLikelihood = 0.0;
			for (unsigned m = 0; m < trainData->getNumSamples(); m++) {
				inputLayer->setVals(trainData->getSample(m));
				for (unsigned k = 0; k < (unsigned)numLevels; k++) {
					unsigned curLevelIdx = numLevels - 1 - k;
					for (unsigned l = 0; l < rectanglesOfLevel[curLevelIdx].size(); l++) {
						PDRectangle *curRectangle = rectanglesOfLevel[curLevelIdx][l];

						for (unsigned n = 0; n < curRectangle->prodNodePtrs_.size(); n++)
							curRectangle->prodNodePtrs_[n]->eval();

						//for (unsigned sumIdx = 0; sumIdx < curRectangle->nodePtrs_.size(); sumIdx++)
						//  curRectangle->nodePtrs_[sumIdx]->eval();

						for (std::map<unsigned, Node* >::iterator it = curRectangle->nodePtrs_.begin();
							it != curRectangle->nodePtrs_.end();
							it++)
							it->second->eval();
					}
				}
				logLikelihood += rectanglesOfLevel[0][0]->nodePtrs_[0]->getLogVal();
			}

			// validation logLikelihood
			valLogLikelihood = 0.0;
			if (options.validationDataSource != NULL) {
				for (unsigned m = 0; m < options.validationDataSource->getNumSamples(); m++) {
					inputLayer->setVals(options.validationDataSource->getSample(m));
					for (unsigned k = 0; k < (unsigned)numLevels; k++) {
						unsigned curLevelIdx = numLevels - 1 - k;
						for (unsigned l = 0; l < rectanglesOfLevel[curLevelIdx].size(); l++) {
							PDRectangle *curRectangle = rectanglesOfLevel[curLevelIdx][l];

							for (unsigned n = 0; n < curRectangle->prodNodePtrs_.size(); n++)
								curRectangle->prodNodePtrs_[n]->eval();

							//for (unsigned sumIdx = 0; sumIdx < curRectangle->nodePtrs_.size(); sumIdx++)
							//  curRectangle->nodePtrs_[sumIdx]->eval();

							for (std::map<unsigned, Node* >::iterator it = curRectangle->nodePtrs_.begin();
								it != curRectangle->nodePtrs_.end();
								it++)
								it->second->eval();
						}
					}
					valLogLikelihood += rectanglesOfLevel[0][0]->nodePtrs_[0]->getLogVal();
				}
			}

			std::cout << "Loglikelihood: Train = " << logLikelihood;
			if (options.validationDataSource != NULL)
				std::cout << "   Val = " << valLogLikelihood << std::endl;
			else
				std::cout << std::endl;

			if (iterC == 0) {
				if (options.validationDataSource != NULL) {
					bestValSPN = convertToNetwork(inputLayer, gaussLayer, rectanglesOfLevel, factory);
					bestValLL = valLogLikelihood;
				}
			}
			else {
				if ((abs(logLikelihood - oldLogLikelihood) < options.stopping_objMinAbsChange) ||
					((logLikelihood - oldLogLikelihood) / abs(logLikelihood) < options.stopping_objMinRelChange)) {
					std::cout << "Loglikelihood increase below threshold, break" << std::endl;
					break;
				}

				if (options.validationDataSource != NULL) {

					if (valLogLikelihood <= bestValLL) 
						numBestValLLnotInc++;
					else {
						for (size_t lc = 0; lc < bestValSPN->getNumLayers(); lc++) {
							for (size_t nc = 0; nc < bestValSPN->getLayer(lc)->getNumNodes(); nc++)
								factory.deleteNode(bestValSPN->getLayer(lc)->getNode(nc));
							factory.deleteLayer(bestValSPN->getLayer(lc));
						}
						factory.deleteNetwork(bestValSPN);
						bestValSPN = convertToNetwork(inputLayer, gaussLayer, rectanglesOfLevel, factory);
						bestValLL = valLogLikelihood;
						numBestValLLnotInc = 0;
					} 						
					
					if (valLogLikelihood <= oldValLogLikelihood)
						numValLLnotInc++;
					else 
						numValLLnotInc = 0;

					if (numValLLnotInc > options.earlyStoppingKrel) {
						std::cout << "ValLoglikelihood has not increased for " << numValLLnotInc << " times, break" << std::endl;
						break;
					}
					if (numBestValLLnotInc > options.earlyStoppingKabs) {
						std::cout << "Best ValLoglikelihood has not increased for " << numBestValLLnotInc << " times, break" << std::endl;
						break;
					}
				}
			}
			
			oldLogLikelihood = logLikelihood;
			oldValLogLikelihood = valLogLikelihood;
		}

		Network* spn;
		if (options.validationDataSource != NULL)
			spn = bestValSPN;
		else
			spn = convertToNetwork(inputLayer, gaussLayer, rectanglesOfLevel, factory);

		// clean up --- free memory
		for (unsigned k = 0; k < rectangles.size(); k++)
			delete rectangles[k];

		for (unsigned k = 0; k < partitions.size(); k++)
			delete partitions[k];

		return spn;
	}


	Network* PDLearning::convertToNetwork(
		ValLayer* inputLayer,
		GaussianLayer* gaussLayer,
		std::map<unsigned, std::vector<PDRectangle*> > &rectanglesOfLevel,
		NetworkFactory &factory) {
		
		size_t numLevels = rectanglesOfLevel.size();
		UMap<NodeId, Node*> nMap;
		Network* spn = factory.generateNetwork();

		// copy input layer
		ValLayer* c_inputLayer = (ValLayer*) factory.generateLayer(Layer::VAL_LAYER, inputLayer->getNumNodes());
		for (size_t k = 0; k < inputLayer->getNumNodes(); k++)
			nMap[inputLayer->getNode(k)->getId()] = c_inputLayer->getNode(k);

		// copy gaussian layer
		GaussianLayer* c_gaussLayer = (GaussianLayer*) factory.generateLayer(Layer::GAUSSIAN_LAYER, gaussLayer->getNumNodes());
		for (size_t k = 0; k < gaussLayer->getNumNodes(); k++) {
			GaussianNode* g = gaussLayer->getNode(k);
			GaussianNode* c_g = c_gaussLayer->getNode(k);
			nMap[g->getId()] = c_g;
			for (size_t c = 0; c < g->getNumChildren(); c++) {
				double mu = ((GaussianNode*)g)->getMean(g->getChild(c));
				double sigma = ((GaussianNode*)g)->getSigma(g->getChild(c));
				c_g->connectChild(nMap.at(g->getChild(c)->getId()));
				((GaussianNode*)c_g)->setMean(nMap.at(g->getChild(c)->getId()), mu);
				((GaussianNode*)c_g)->setSigma(nMap.at(g->getChild(c)->getId()), sigma);
			}
		}
		
		spn->addLayer(c_inputLayer);
		spn->addLayer(c_gaussLayer);

		for (unsigned k = 0; k < (unsigned)numLevels; k++) {
			unsigned curLevelIdx = (unsigned)numLevels - 1 - k;
			
			SumLayer* sumLayer = (SumLayer*)factory.generateLayer(Layer::SUM_LAYER, 0);
			ProdLayer* prodLayer = (ProdLayer*)factory.generateLayer(Layer::PROD_LAYER, 0);

			for (unsigned l = 0; l < rectanglesOfLevel[curLevelIdx].size(); l++) {
				PDRectangle *curRectangle = rectanglesOfLevel[curLevelIdx][l];

				if ((curRectangle->width_ == 1) && (curRectangle->height_ == 1))
					continue;

				for (unsigned pC = 0; pC < curRectangle->prodNodePtrs_.size(); pC++) {
					Node *prod = curRectangle->prodNodePtrs_[pC];
					Node *c_prod = factory.generateNode(Node::PROD_NODE);
					nMap[prod->getId()] = c_prod;

					for (size_t c = 0; c < prod->getNumChildren(); c++) {
						Node* child = prod->getChild(c);
						c_prod->connectChild(nMap.at(child->getId()));
					}

					prodLayer->insertNode(c_prod);
				}

				for (auto it = curRectangle->nodePtrs_.begin(); it != curRectangle->nodePtrs_.end(); it++) {
					if (it->second->getType() != Node::SUM_NODE)
						throw std::runtime_error("PDLearn: not a sum in non-unit rectangle");

					SumNode* sum = (SumNode*) it->second;
					SumNode* c_sum = (SumNode*) factory.generateNode(Node::SUM_NODE);
					nMap[sum->getId()] = c_sum;

					for (size_t c = 0; c < sum->getNumChildren(); c++) {
						Node* child = sum->getChild(c);
						c_sum->connectChild(nMap.at(child->getId()));
						c_sum->setLogWeight(nMap.at(child->getId()), sum->getLogWeight(child));
					}

					sumLayer->insertNode(c_sum);
				}
			}

			spn->addLayer(prodLayer);
			spn->addLayer(sumLayer);
		}
		

		//
		// prune 0-weights
		//
		for (size_t k = 0; k < spn->getNumLayers(); k++) {
			if (spn->getLayer(k)->getType() != Layer::SUM_LAYER)
				continue;
			for (unsigned l = 0; l < spn->getLayer(k)->getNumNodes(); l++) {
				Node* node = spn->getLayer(k)->getNode(l);

				NodeVec nodesToDisconnect;
				for (unsigned c = 0; c < node->getNumChildren(); c++)
					if (((SumNode*)node)->getWeight(node->getChild(c)) < numericZero)
						nodesToDisconnect.push_back(node->getChild(c));
								
				for (unsigned c = 0; c < nodesToDisconnect.size(); c++)
					node->disconnectChild(nodesToDisconnect[c]);
			}
		}

		//
		// top -> bottom: remove parent-less nodes
		//
		for (size_t k = 1; k < spn->getNumLayers(); k++) {
			size_t idx = spn->getNumLayers() - 1 - k;
			NodeVec nodesToRemove;
			for (unsigned l = 0; l < spn->getLayer(idx)->getNumNodes(); l++) {				
				Node* node = spn->getLayer(idx)->getNode(l);

				if (node->getNumParents() == 0) 
					nodesToRemove.push_back(node);

				if ((node->getType() == Node::VAL_NODE) && (node->getNumParents() == 0))
					throw std::runtime_error("PDLearn: valnode without parents");
			}
			spn->getLayer(idx)->removeNodes(nodesToRemove);
			for (unsigned l = 0; l < nodesToRemove.size(); l++)
				factory.deleteNode(nodesToRemove[l]);
		}

		//
		// bottom -> top: remove child-less nodes
		//
		for (unsigned k = 1; k < spn->getNumLayers(); k++) {
			unsigned idx = k;
			NodeVec nodesToRemove;
			for (unsigned l = 0; l < spn->getLayer(idx)->getNumNodes(); l++) {
				Node* node = spn->getLayer(idx)->getNode(l);

				if (node->getNumChildren() == 0)
					nodesToRemove.push_back(node);
			}
			spn->getLayer(idx)->removeNodes(nodesToRemove);
			for (unsigned l = 0; l < nodesToRemove.size(); l++)
				factory.deleteNode(nodesToRemove[l]);
		}


		LayerVec LayersToRemove;
		for (unsigned k = 0; k < spn->getNumLayers(); k++)
			if (spn->getLayer(k)->getNumNodes() == 0)
				LayersToRemove.push_back(spn->getLayer(k));
		for (unsigned k = 0; k < LayersToRemove.size(); k++)
			spn->removeLayer(LayersToRemove[k]);

		
		// assert
		if (spn->getLayer(0)->getNumNodes() != inputLayer->getNumNodes())
			throw std::runtime_error("PDLearn: input layer has too few nodes");
				
		for (size_t k = 0; k < spn->getNumLayers(); k++) {
			if (spn->getLayer(k)->getNumNodes() == 0)
				throw std::runtime_error("PDLearn: empty layer");
			for (size_t l = 0; l < spn->getLayer(k)->getNumNodes(); l++) {
				if ((spn->getLayer(k)->getNode(l)->getNumParents() == 0) && (k + 1 < spn->getNumLayers())) {
					std::cout << spn->getNumLayers() << " " << k << " " << spn->getLayer(k)->getNode(l)->getType() << std::endl;
					throw std::runtime_error("PDLearn: parent-less node");
				}
				if ((spn->getLayer(k)->getNode(l)->getNumChildren() == 0) && (spn->getLayer(k)->getNode(l)->getType() != Node::VAL_NODE))
					throw std::runtime_error("PDLearn: child-less node");
			}
		}

		return spn;
	}

}
