/*
 *
 * Implementation of the PD learning algorithms:
 * http://alchemy.cs.washington.edu/spn/
 *
 */

#ifndef PDLEARNING_HPP
#define	PDLEARNING_HPP

#include "Options.hpp"
#include "TypeDefs.hpp"
#include "NetworkFactory.hpp"
#include "StatsCounterFloat.hpp"
#include "GraphTools.hpp"
#include "Utils.hpp"

namespace SPN {

	class PDRectangle;
	class PDPartition;

	typedef std::vector<PDRectangle*> PDRectangleVec;
	typedef std::vector<PDPartition*> PDPartitionVec;

	class PDRectangle {
	public:
		unsigned x_, y_, width_, height_;
		int level_;
		NodeVec prodNodePtrs_;
		std::map<unsigned, Node* > nodePtrs_;
		PDPartitionVec parentPartitions_;
		PDPartitionVec childPartitions_;

		unsigned MAPNodeIdx_;
		double MAPNodeLogVal_;

		PDRectangle();
		void setLevelRecursive(int level);

		bool operator<(const PDRectangle &rhs) const;
		bool operator==(const PDRectangle &rhs) const;
		bool operator!=(const PDRectangle &rhs) const;
	};

	class PDPartition {
	public:
		PDRectangle* parentRectangle_;
		PDRectangleVec childRectangles_;

		PDPartition();
	};

	class PDProdNodeId {
		friend class PDLearning;
	private:
		unsigned x1_;
		unsigned y1_;
		unsigned width1_;
		unsigned height1_;
		unsigned x2_;
		unsigned y2_;
		unsigned width2_;
		unsigned height2_;
		unsigned idx1_;
		unsigned idx2_;
	public:
		PDRectangle getRectangle1();
		unsigned getIdx1();
		PDRectangle getRectangle2();
		unsigned getIdx2();
		void set(PDRectangle &r1, PDRectangle &r2, unsigned idx1, unsigned idx2);
		PDProdNodeId(PDRectangle &r1, PDRectangle &r2, unsigned idx1, unsigned idx2);
		PDProdNodeId();
		bool operator<(const PDProdNodeId &rhs) const;
		bool operator==(const PDProdNodeId &rhs) const;
		bool operator!=(const PDProdNodeId &rhs) const;
	};



	class PDLearning {
	private:

	protected:
		int pixelOrder_;
		int width_;
		int height_;
		int numChannels_;
		int numGaussPerPixel_;
		int maxNumSumsPerRectangle_;
		UnsignedVec coarseRes_;
		UnsignedVec coarseResX_;
		UnsignedVec coarseResY_;
		bool preSubtractCounts_;
		bool flatEM_;
		DoubleVec2 gaussMeans_;

		unsigned getInputIdx(int x, int y, int c);
		bool incCounter(UnsignedVec &counter, UnsignedVec &maxVals);
		void assertRectangles(PDRectangle *parent, PDRectangle *child1, PDRectangle *child2);

		void setWeights(PDRectangleVec &rectangles, const StatsCounterFloat &counts);

		Network* convertToNetwork(
			ValLayer* inputLayer,
			GaussianLayer* gaussLayer,
			std::map<unsigned, std::vector<PDRectangle*> > &rectanglesOfLevel,
			NetworkFactory &factory);


	public:
		static const int ROW_MAJOR = 1;
		static const int COLUMN_MAJOR = 2;
		static const int GRAY = 1;
		static const int RGB = 3;

		Options options;

		PDLearning();
		virtual ~PDLearning();

		// how are the pixels arranged in the linear memory
		void setRowMajor();
		void setColumnMajor();

		// set the number of color channels
		void setNumChannels(int numChannels);

		// set width, height of input image
		void setDims(int width, int height);
		void setWidth(int width);
		void setHeight(int width);

		// maximal number of sum nodes per rectangle
		void setMaxNumSumsPerRectangle(int maxNumSums);

		// number of Gaussians per pixel
		void setNumGaussPerPixel(int numGauss);

		// coarse resolutions --- 4 in the PD code
		void setCoarseRes(unsigned coarseRes);
		void setCoarseRes(UnsignedVec coarseRes);

		// coarse resolutions for X 
		void setCoarseResX(unsigned coarseResX);
		void setCoarseResX(UnsignedVec coarseResX);

		// coarse resolutions for Y 
		void setCoarseResY(unsigned coarseResY);
		void setCoarseResY(UnsignedVec coarseResY);

		// set pre-defined Gaussian means
		void setGaussMeans(DoubleVec2 means);

		// in the PD code, the old batch statistics are subtracted before inference.
		// To use the original version, set to preSubtractCounts_ = true.
		void preSubtract(bool flag) {
			preSubtractCounts_ = flag;
		}

		// In the PD code, the current child counts are not considered, when the winning child is selected.
		// To run the original version set flatEM_ = true.
		void flatEM(bool flag) {
			flatEM_ = flag;
		}

		// 
		Network* learn(DataSource* trainData, NetworkFactory &factory);
	};
}

#endif	/* PDLEARNING_HPP */

