
#include "StatsCounterFloat.hpp"

namespace SPN {

	StatsCounterFloat::StatsCounterFloat() {
		clear();
	}

	StatsCounterFloat::~StatsCounterFloat() {
	}

	void StatsCounterFloat::clear() {
		counts_.clear();
		childCounts_.clear();
		numNonzeroChildren_.clear();
	}

	void StatsCounterFloat::add(const Node& sumNode, const Node& childNode, float count) {
		if (sumNode.getType() != Node::SUM_NODE)
			throw std::logic_error("StatsCounterFloat::add(Node& sumNode, Node& childNode, float count):"
			"sumNode not of type SumNode");

		NodeId sumId = sumNode.getId();
		bool childWasZero = childCounts_[sumId][childNode.getId()] <= numericZero;

		childCounts_[sumId][childNode.getId()] += count;
		counts_[sumId] += count;

		if (childCounts_[sumId][childNode.getId()] < -numericZero)
			throw std::logic_error("StatsCounterFloat::add(Node& sumNode, Node& childNode, float count):"
			"negative counts");

		if ((childWasZero) && (counts_[sumId] > numericZero))
			numNonzeroChildren_[sumId]++;

		if ((!childWasZero) && (counts_[sumId] <= numericZero))
			numNonzeroChildren_[sumId]--;
	}

	void StatsCounterFloat::subtract(const Node& sumNode, const Node& childNode, float count) {
		if (sumNode.getType() != Node::SUM_NODE)
			throw std::logic_error("StatsCounterFloat::subtract(Node& sumNode, Node& childNode, float count):"
			"sumNode not of type SumNode");

		NodeId sumId = sumNode.getId();
		bool childWasZero = childCounts_[sumId][childNode.getId()] <= numericZero;

		childCounts_[sumId][childNode.getId()] -= count;
		counts_[sumId] -= count;

		if (childCounts_[sumId][childNode.getId()] < -numericZero)
			throw std::logic_error("StatsCounterFloat::subtract(Node& sumNode, Node& childNode, float count):"
			"negative counts");

		if ((childWasZero) && (counts_[sumId] > numericZero))
			numNonzeroChildren_[sumId]++;

		if ((!childWasZero) && (counts_[sumId] <= numericZero))
			numNonzeroChildren_[sumId]--;
	}

	void StatsCounterFloat::set(const Node& sumNode, const Node& childNode, float count) {
		if (sumNode.getType() != Node::SUM_NODE)
			throw std::logic_error("StatsCounterFloat::set(Node& sumNode, Node& childNode, float count):"
			"sumNode not of type SumNode");

		NodeId sumId = sumNode.getId();
		NodeId childId = childNode.getId();
		bool childWasZero = childCounts_[sumId][childId] <= numericZero;

		counts_[sumId] += count - childCounts_[sumId][childId];
		childCounts_[sumId][childId] = count;

		if (childCounts_[sumId][childNode.getId()] < -numericZero)
			throw std::logic_error("StatsCounterFloat::set(Node& sumNode, Node& childNode, float count):"
			"negative counts");

		if ((childWasZero) && (counts_[sumId] > numericZero))
			numNonzeroChildren_[sumId]++;

		if ((!childWasZero) && (counts_[sumId] <= numericZero))
			numNonzeroChildren_[sumId]--;
	}

	float StatsCounterFloat::get(const Node& sumNode, const Node& childNode) const {
		if (sumNode.getType() != Node::SUM_NODE)
			throw std::logic_error("StatsCounterFloat::set(Node& sumNode, Node& childNode, float count):"
			"sumNode not of type SumNode");

		NodeId sumId = sumNode.getId();
		NodeId childId = childNode.getId();
		if ((childCounts_.count(sumId) == 0) || (childCounts_.at(sumId).count(childId) == 0))
			return 0.0;
		return childCounts_.at(sumId).at(childId);
	}

	float StatsCounterFloat::get(const Node& sumNode) const {
		if (sumNode.getType() != Node::SUM_NODE)
			throw std::logic_error("StatsCounterFloat::set(Node& sumNode, Node& childNode, float count):"
			"sumNode not of type SumNode");

		NodeId sumId = sumNode.getId();
		if (counts_.count(sumId) == 0)
			return 0.0;
		return counts_.at(sumId);
	}

	StatsCounterFloat& StatsCounterFloat::operator+=(StatsCounterFloat const& rhs) {
		
		for (auto countsIt = rhs.counts_.begin(); countsIt != rhs.counts_.end(); countsIt++)
			counts_[countsIt->first] += countsIt->second;

		for (auto childCountsIt = rhs.childCounts_.begin();
			childCountsIt != rhs.childCounts_.end();
			childCountsIt++) {
			for (auto countsIt = childCountsIt->second.begin();
				countsIt != childCountsIt->second.end();
				countsIt++) {
				bool childWasZero = childCounts_[childCountsIt->first][countsIt->first] <= numericZero;

				childCounts_[childCountsIt->first][countsIt->first] += countsIt->second;

				if ((childWasZero) && (childCounts_[childCountsIt->first][countsIt->first] > numericZero))
					numNonzeroChildren_[childCountsIt->first]++;

				if ((!childWasZero) && (childCounts_[childCountsIt->first][countsIt->first] <= numericZero))
					numNonzeroChildren_[childCountsIt->first]--;
			}
		}
		return *this;
	}

	StatsCounterFloat& StatsCounterFloat::operator-=(StatsCounterFloat const& rhs) {
		
		for (auto countsIt = rhs.counts_.begin(); countsIt != rhs.counts_.end(); countsIt++)
			counts_[countsIt->first] -= countsIt->second;

		for (auto childCountsIt = rhs.childCounts_.begin();
			childCountsIt != rhs.childCounts_.end();
			childCountsIt++) {
			for (auto countsIt = childCountsIt->second.begin();
				countsIt != childCountsIt->second.end();
				countsIt++) {
				bool childWasZero = childCounts_[childCountsIt->first][countsIt->first] <= numericZero;

				childCounts_[childCountsIt->first][countsIt->first] -= countsIt->second;

				if ((childWasZero) && (childCounts_[childCountsIt->first][countsIt->first] > numericZero))
					numNonzeroChildren_[childCountsIt->first]++;

				if ((!childWasZero) && (childCounts_[childCountsIt->first][countsIt->first] <= numericZero))
					numNonzeroChildren_[childCountsIt->first]--;
			}
		}
		return *this;
	}

	unsigned StatsCounterFloat::getNumNonzeroChildren(const Node& sumNode) const {
		if (sumNode.getType() != Node::SUM_NODE)
			throw std::logic_error("StatsCounterFloat::getNumNonzeroChildren(const Node& sumNode) const:"
			"sumNode not of type SumNode");

		NodeId sumId = sumNode.getId();
		if (numNonzeroChildren_.count(sumId) == 0)
			return 0;
		return numNonzeroChildren_.at(sumId);
	}

	size_t StatsCounterFloat::getNumEntries() const {
		size_t ret = counts_.size();				
		for (auto it = childCounts_.begin(); it != childCounts_.end(); it++)
			ret += it->second.size();
		return ret;
	}
}
