
#include "Node.hpp"

namespace SPN {

	SumNode::SumNode() : Node(SUM_NODE) {
		logVal_ = -inf;
		maxLogVal_ = -inf;
		maxIdx_ = 0;
		weights_.clear();
		logWeights_.clear();
		maxMode_ = false;
	}

	SumNode::~SumNode() {
	}

	void SumNode::eval() {
		size_t numChildren = getNumChildren();

		if (numChildren == 0) {
			maxIdx_ = 0;
			maxLogVal_ = -inf;
			logVal_ = -inf;
			return;
		}		

		double val = 0.0;
		double *computeCache = new double[weights_.size()];

		maxLogVal_ = -inf;
		maxIdx_ = 0;

		for (unsigned k = 0; k < numChildren; k++) {
			Node* child = getChild(k);
			NodeId childId = child->getId();
			computeCache[k] = child->getLogVal() + logWeights_[childId];
			if (computeCache[k] > maxLogVal_) {
				maxLogVal_ = computeCache[k];
				maxIdx_ = k;
			}
		}

		if (maxLogVal_ > -inf) {
			for (unsigned k = 0; k < numChildren; k++)
				val += exp(computeCache[k] - maxLogVal_);
		}
		else
			val = 0.0;

		if (isnan(val))
			logVal_ = nan;
		else if (val > 0)
			logVal_ = maxLogVal_ + log(val);
		else
			logVal_ = -inf;

		delete[] computeCache;
	}

	double SumNode::getLogVal() const {
		if (maxMode_)
			return maxLogVal_;
		else
			return logVal_;
	}

	double SumNode::getVal() const {
		if (maxMode_)
			return exp(maxLogVal_);
		else
			return exp(logVal_);
	}

	unsigned SumNode::getMaxIdx() const {
		return maxIdx_;
	}

	double SumNode::getLog_Dval_Dchild(Node *child) {
		if (!isChild(child))
			throw std::runtime_error("SumNode::getLog_Dval_Dchild(Node *child): child node not found");

		NodeId childId = child->getId();
		if (logWeights_.count(childId) == 0)
			throw std::runtime_error("SumNode::getLog_Dval_Dchild(Node *child): child weight not found");
		return logWeights_[childId];
	}

	void SumNode::connectChild(Node *child) {
		connectChild(child, 0.0);
	}

	void SumNode::connectChild(Node *child, double weight) {
		if ((isChild(child)) || (child->getId() == getId()))
			return;

		Node::connectChild(child);
		setWeight(child, weight);
	}

	void SumNode::disconnectChild(Node *child) {
		Node::disconnectChild(child);
		NodeId childId = child->getId();
		weights_.erase(childId);
		logWeights_.erase(childId);
	}

	void SumNode::setWeight(Node* child, double weight) {
		if (!isChild(child))
			throw std::runtime_error("error in SumNode::setWeight(Node *child, double weight): not a child.");
		if (weight < 0.0)
			throw std::runtime_error("error in SumNode::setWeight(Node *child, double weight): negative weight.");

		NodeId childId = child->getId();
		weights_[childId] = weight;
		if (weight > 0)
			logWeights_[childId] = log(weight);
		else
			logWeights_[childId] = -inf;
	}

	void SumNode::setWeights(const DoubleVec &weights) {
		size_t numChildren = getNumChildren();

		if (numChildren != weights.size())
			throw std::runtime_error("error in SumNode::setWeights(DoubleVec weights): "
			"weights vector must contain getNumChildren() entries.");

		for (unsigned k = 0; k < numChildren; k++)
			setWeight(getChild(k), weights[k]);
	}

	void SumNode::setLogWeight(Node *child, double logWeight) {
		if (!isChild(child))
			throw std::runtime_error("error in SumNode::setLogWeight(Node *child, double weight): not a child.");

		size_t childId = child->getId();
		logWeights_[childId] = logWeight;
		weights_[childId] = exp(logWeight);
	}

	void SumNode::setLogWeights(const DoubleVec &logWeights) {
		size_t numChildren = getNumChildren();

		if (numChildren != logWeights.size())
			throw std::runtime_error("error in SumNode::setLogWeights(DoubleVec logWeights): "
			"weights vector must contain getNumChildren() entries.");

		for (unsigned k = 0; k < numChildren; k++)
			setLogWeight(getChild(k), logWeights[k]);
	}

	double SumNode::getWeight(Node *child) {
		if (!isChild(child))
			throw std::runtime_error("SumNode::getWeight(Node *child): not a child.");
		return weights_[child->getId()];
	}

	double SumNode::getLogWeight(Node *child) {
		if (!isChild(child))
			throw std::runtime_error("SumNode::getLogWeight(Node *child): not a child.");
		return logWeights_[child->getId()];
	}

	// stupid way to draw from Dirichlet distribution with uniform alpha=1
	void SumNode::initWeightsRand() {
		double sum = 0.0;
		DoubleVec r;

		r.resize(getNumChildren());
		for (unsigned k = 0; k < getNumChildren(); k++) {

			double gsample1 = 0.0;			
			for (unsigned l = 0; l < 48;l ++)
				gsample1 += ((double)rand() / double(RAND_MAX)) - 0.5;
			gsample1 = gsample1 * 0.5;

			double gsample2 = 0.0;
			for (unsigned l = 0; l < 48; l++)
				gsample2 += ((double)rand() / double(RAND_MAX)) - 0.5;
			gsample2 = gsample2 * 0.5;
			
			r[k] = gsample1 * gsample1 + gsample2 * gsample2;
			sum += r[k];
		}
		for (unsigned k = 0; k < weights_.size(); k++)
			r[k] = r[k] / sum;

		setWeights(r);

		//std::cout << vec2string(r) << std::endl;
	}

	void SumNode::normalizeWeights() {
		double sum = 0.0;
		DoubleVec r;
		r.resize(getNumChildren());

		for (unsigned k = 0; k < getNumChildren(); k++) {
			r[k] = getWeight(getChild(k));
			sum += r[k];
		}

		if (sum == 0)
			throw std::runtime_error("SumNode::normalizeWeights(): sum of weights == 0");

		for (unsigned k = 0; k < getNumChildren(); k++)
			setWeight(getChild(k), r[k] / sum);
	}

	void SumNode::setSumMode() {
		maxMode_ = false;
	}

	void SumNode::setMaxMode() {
		maxMode_ = true;
	}

}
